#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint:disable=import-error
"""Intel Neural Compressor Pytorch quantization config API."""


import copy
import inspect
import json
from collections import OrderedDict
from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Optional
from typing import OrderedDict as OrderedDictType
from typing import Tuple, Union

import torch

import neural_compressor.torch.utils as torch_utils
from neural_compressor.common.base_config import (
    BaseConfig,
    config_registry,
    register_config,
    register_supported_configs_for_fwk,
)
from neural_compressor.common.utils import (
    AUTOROUND,
    AWQ,
    DEFAULT_WHITE_LIST,
    FP8_QUANT,
    GPTQ,
    HQQ,
    HYBRID_GPTQ,
    MIXED_PRECISION,
    MX_QUANT,
    OP_NAME_OR_MODULE_TYPE,
    RTN,
    SMOOTH_QUANT,
    STATIC_QUANT,
    TEQ,
)
from neural_compressor.torch.utils import is_hpex_available, is_ipex_imported, is_transformers_imported, logger
from neural_compressor.torch.utils.constants import (
    LM_HEAD_NAMES,
    PRIORITY_AUTOROUND,
    PRIORITY_AWQ,
    PRIORITY_GPTQ,
    PRIORITY_HQQ,
    PRIORITY_RTN,
    PRIORITY_TEQ,
    PT2E_DYNAMIC_QUANT,
)

FRAMEWORK_NAME = "torch"
if is_transformers_imported():
    import transformers

    WOQ_WHITE_LIST = (torch.nn.Linear, transformers.Conv1D)
else:
    WOQ_WHITE_LIST = (torch.nn.Linear,)


class OperatorConfig(NamedTuple):
    """OperatorConfig."""

    config: BaseConfig
    operators: List[Union[str, Callable]]
    valid_func_list: List[Callable] = []


class TorchBaseConfig(BaseConfig):
    """Base config class for torch backend."""

    # re-write func _get_op_name_op_type_config to fallback op_type with string
    # because there are some special op_types for IPEX backend: `Linear&Relu`, `Linear&add`, ...
    def __init__(self, white_list=DEFAULT_WHITE_LIST):
        super().__init__(white_list)
        self.params_list = self.__class__._generate_params_list()
        self.non_tunable_params: List[str] = ["white_list"]

    def _get_op_name_op_type_config(self):
        op_type_config_dict = dict()
        op_name_config_dict = dict()
        for name, config in self.local_config.items():
            if self._is_op_type(name):
                # Convert the Callable to String.
                new_name = self._op_type_to_str(name)
                op_type_config_dict[new_name] = config
            else:
                op_name_config_dict[name] = config
                if is_ipex_imported():
                    op_type_config_dict[name] = config
        return op_type_config_dict, op_name_config_dict

    @classmethod
    def _generate_params_list(cls) -> List[str]:
        sig = inspect.signature(cls.__init__)
        params_list = list(sig.parameters.keys())[1:]
        if "args" in params_list:
            params_list.remove("args")
        if "kwargs" in params_list:
            params_list.remove("kwargs")
        return params_list


######################## RNT Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=RTN, priority=PRIORITY_RTN)
class RTNConfig(TorchBaseConfig):
    """Config class for round-to-nearest weight-only quantization."""

    name = RTN
    supported_configs: List[OperatorConfig] = []

    def __init__(
        self,
        dtype: str = "int",
        bits: int = 4,
        use_sym: bool = True,
        group_size: int = 32,
        group_dim: int = 1,
        use_full_range: bool = False,
        use_mse_search: bool = False,
        # layer wise
        use_layer_wise: bool = False,
        model_path: str = "",
        # double quant
        use_double_quant: bool = False,
        double_quant_dtype: str = "int",
        double_quant_bits: int = 8,  # not available when double_quant_dtype is not 'int'
        double_quant_use_sym: bool = False,
        double_quant_group_size: int = 256,
        # quant lm_head
        quant_lm_head: bool = False,
        # Tuning space
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init RTN weight-only quantization config.

        Args:
            dtype (str): Data type for weights. Default is "int".
            bits (int): Number of bits used to represent weights. Default is 4.
            use_sym (bool): Indicates whether weights are symmetric. Default is True.
            group_size (int): Size of weight groups. Default is 32.
            group_dim (int): Dimension for grouping. Default is 1.
            use_full_range (bool): Enables full range for activations. Default is False.
            use_mse_search (bool): Enables mean squared error (MSE) search. Default is False.
            use_layer_wise (bool): Enables quantize model per layer. Defaults to False.
            model_path (str): Model path that is used to load state_dict per layer.
            use_double_quant (bool): Enables double quantization. Default is False.
            double_quant_dtype (str): Data type for double_quant scale. Default is "int".
            double_quant_bits (int): Number of bits used to represent double_quant scale. Default is 4.
            double_quant_use_sym (bool): Indicates whether double_quant scale are symmetric. Default is True.
            double_quant_group_size (int): Size of double_quant groups. Default is 32.
            quant_lm_head (bool): Indicates whether quantize the lm_head layer in transformers。 Default is False.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
                Default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.dtype = dtype
        self.bits = bits
        self.use_sym = use_sym
        self.group_size = group_size
        self.group_dim = group_dim
        self.use_full_range = use_full_range
        self.use_mse_search = use_mse_search
        self.use_layer_wise = use_layer_wise
        self.model_path = model_path
        # double quant
        self.use_double_quant = use_double_quant
        self.double_quant_bits = double_quant_bits
        self.double_quant_dtype = double_quant_dtype
        self.double_quant_use_sym = double_quant_use_sym
        self.double_quant_group_size = double_quant_group_size
        self.quant_lm_head = quant_lm_head
        self._post_init()  # initialize global & local configuration

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations for RTN.

        Returns:
            List[OperatorConfig]: List of supported operator configurations.
        """
        supported_configs = []
        linear_rtn_config = RTNConfig(
            dtype=[
                "int",
                "int8",
                "int4",
                "nf4",
                "fp4",
                "fp4_e2m1_bnb",
                "fp4_e2m1",
                "fp8_e5m2",
                "fp8_e5m2fnuz",
                "fp8_e4m3fn",
                "fp8_e4m3fnuz",
            ],
            bits=[4, 1, 2, 3, 5, 6, 7, 8],
            use_sym=[True, False],
            group_size=[32, -1, 1, 4, 8, 16, 64, 128, 256, 512, 1024],
            group_dim=[1, 0],
            use_full_range=[False, True],
            use_mse_search=[False, True],
            use_layer_wise=[False, True],
            use_double_quant=[False, True],
            double_quant_bits=[4, 1, 2, 3, 5, 6, 7, 8],
            double_quant_dtype=["int"],
            double_quant_use_sym=[True, False],
            double_quant_group_size=[32, -1, 1, 4, 8, 16, 64, 128, 256, 512, 1024],
            quant_lm_head=[False, True],
        )
        operators = list(WOQ_WHITE_LIST)
        supported_configs.append(OperatorConfig(config=linear_rtn_config, operators=operators))
        cls.supported_configs = supported_configs

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        if not self.quant_lm_head:
            self.set_local(
                LM_HEAD_NAMES, RTNConfig(dtype="fp32", use_layer_wise=self.use_layer_wise, model_path=self.model_path)
            )
        config_mapping = super().to_config_mapping(config_list, model_info)
        return config_mapping

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        filter_result = []
        for op_name, module in model.named_modules():
            if isinstance(module, WOQ_WHITE_LIST):
                pair = (op_name, type(module).__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "RTNConfig", List["RTNConfig"]]:
        """Get the default configuration set for tuning.

        Returns:
            Union[None, "RTNConfig", List["RTNConfig"]]: The configuration set for tuning.
        """
        return RTNConfig(
            dtype=["int4", "nf4"], use_sym=[True, False], group_size=[32, 128], use_mse_search=[False, True]
        )

    @classmethod
    def get_predefined_configs(cls) -> Dict[torch_utils.ProcessorType, "RTNConfig"]:
        """Get the predefined configuration set.

        Returns:
            Dict[torch_utils.ProcessorType, "RTNConfig"]: The configuration of RTN.
        """
        pre_defined_configs: Dict[torch_utils.ProcessorType, RTNConfig] = {}
        pre_defined_configs[torch_utils.ProcessorType.Client] = cls(use_layer_wise=True)
        pre_defined_configs[torch_utils.ProcessorType.Server] = cls()
        return pre_defined_configs


def get_default_rtn_config(processor_type: Optional[Union[str, torch_utils.ProcessorType]] = None) -> RTNConfig:
    """Get the default configuration of RTN.

    Args:
        processor_type (Optional[Union[str, torch_utils.ProcessorType]], optional): The user-specified processor type.
            Defaults to None.

    Returns:
        RTNConfig: RTNConfig
    """
    process_type = torch_utils.get_processor_type_from_user_config(processor_type)
    return RTNConfig.get_predefined_configs()[process_type]


def get_default_double_quant_config(type="BNB_NF4"):
    """Get the default configuration of double quant.

    Args:
        type (str, optional): double quant type. Defaults to "BNB_NF4".

    Returns:
        dict: double quant config.
    """
    from neural_compressor.torch.utils.constants import DOUBLE_QUANT_CONFIGS

    assert type in DOUBLE_QUANT_CONFIGS, "Supported double quant configs: {}".format(list(DOUBLE_QUANT_CONFIGS.keys()))
    return RTNConfig.from_dict(DOUBLE_QUANT_CONFIGS[type])


######################## GPTQ Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=GPTQ, priority=PRIORITY_GPTQ)
class GPTQConfig(TorchBaseConfig):
    """Config class for GPTQ.

    GPTQ: Accurate Post-Training Quantization for Generative Pre-trained Transformers.
    https://arxiv.org/abs/2210.17323
    """

    name = GPTQ
    supported_configs: List[OperatorConfig] = []

    def __init__(
        self,
        dtype: str = "int",
        bits: int = 4,
        use_sym: bool = True,
        group_size: int = 32,
        use_mse_search: bool = False,
        # layer wise
        use_layer_wise: bool = False,
        use_block_wise: bool = False,
        model_path: str = "",
        # double quant
        use_double_quant: bool = False,
        double_quant_dtype: str = "int",
        double_quant_bits: int = 8,  # not available when double_quant_dtype is not 'int'
        double_quant_use_sym: bool = False,
        double_quant_group_size: int = 256,
        # double quant
        quant_lm_head: bool = False,
        # gptq params
        act_order: bool = False,
        hybrid_order: bool = False,
        fp8_aware: bool = False,
        percdamp: float = 0.01,
        block_size: int = 2048,
        static_groups: bool = False,
        true_sequential: bool = False,
        # Tuning space
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init GPTQ weight-only quantization config.

        Args:
            dtype (str): Data type for weights. Default is "int".
            bits (int): Number of bits used to represent weights. Default is 4.
            use_sym (bool): Indicates whether weights are symmetric. Default is True.
            group_size (int): Size of weight groups. Default is 32.
            use_mse_search (bool): Enables mean squared error (MSE) search. Default is False.
            use_layer_wise (bool): Enables quantize model per layer. Defaults to False.
            use_block_wise (bool): Enables quantize model per block. Defaults to False. Used for large GPU/HPU.
            model_path (str): Model path that is used to load state_dict per layer.
            use_double_quant (bool): Enables double quantization. Default is False.
            double_quant_dtype (str): Data type for double_quant scale. Default is "int".
            double_quant_bits (int): Number of bits used to represent double_quant scale. Default is 4.
            double_quant_use_sym (bool): Indicates whether double_quant scale are symmetric. Default is True.
            double_quant_group_size (int): Size of double_quant groups. Default is 32.
            quant_lm_head (bool): Indicates whether quantize the lm_head layer in transformers。 Default is False.
            act_order (bool): Whether to sort Hessian's diagonal values to rearrange channel-wise
                              quantization order. Default is False.
            hybrid_order (bool): Enables activation re-ordering with no inference overhead.
                                 Weights are re-ordered within their groups without cross-group mixing.
            fp8_aware (bool): Whether to include an FP8 quantization step in the GPTQ process.
                              This improves accuracy when using the W4A8 quantization scheme.
            percdamp (float): Percentage of Hessian's diagonal values' average, which will be added to
                              Hessian's diagonal to increase numerical stability. Default is 0.01.
            block_size (int): Execute GPTQ quantization per block, block shape = [C_out, block_size].
                              Default is 128.
            static_groups (bool): Whether to calculate group wise quantization parameters in advance.
                                  This option mitigate actorder's extra computational requirements.
                                  Default is False.
            true_sequential (bool): Whether to quantize layers within a transformer block in their original order.
                                  This can lead to higher accuracy but slower overall quantization process.
                                  Default is False.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
                                                                 Default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.dtype = dtype
        self.bits = bits
        self.use_sym = use_sym
        self.group_size = group_size
        self.use_mse_search = use_mse_search
        # layer wise
        self.use_layer_wise = use_layer_wise
        self.use_block_wise = use_block_wise
        self.model_path = model_path
        # double quant
        self.use_double_quant = use_double_quant
        self.double_quant_bits = double_quant_bits
        self.double_quant_dtype = double_quant_dtype
        self.double_quant_use_sym = double_quant_use_sym
        self.double_quant_group_size = double_quant_group_size
        # gptq
        self.act_order = act_order
        self.hybrid_order = hybrid_order
        self.fp8_aware = fp8_aware
        self.percdamp = percdamp
        self.block_size = block_size
        self.static_groups = static_groups
        self.true_sequential = true_sequential
        self.quant_lm_head = quant_lm_head
        self._post_init()  # initialize global & local configuration

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations for GPTQ.

        Returns:
            List[OperatorConfig]: List of supported operator configurations.
        """
        supported_configs = []
        # TODO(Yi)
        linear_gptq_config = GPTQConfig()
        operators = list(WOQ_WHITE_LIST)
        supported_configs.append(OperatorConfig(config=linear_gptq_config, operators=operators))
        cls.supported_configs = supported_configs

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        if not self.quant_lm_head:
            self.set_local(
                LM_HEAD_NAMES,
                GPTQConfig(
                    dtype="fp32",
                    use_layer_wise=self.use_layer_wise,
                    model_path=self.model_path,
                    use_block_wise=self.use_block_wise,
                ),
            )
        config_mapping = super().to_config_mapping(config_list, model_info)
        return config_mapping

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        filter_result = []
        for op_name, module in model.named_modules():
            if isinstance(module, WOQ_WHITE_LIST):
                pair = (op_name, type(module).__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "GPTQConfig", List["GPTQConfig"]]:
        """Get the default configuration set for tuning.

        Returns:
            Union[None, "GPTQConfig", List["GPTQConfig"]]: The configuration set for tuning.
        """
        # TODO fwk owner needs to update it.
        return GPTQConfig(act_order=[True, False], use_sym=[False, True])

    @classmethod
    def get_predefined_configs(cls) -> Dict[torch_utils.ProcessorType, "GPTQConfig"]:
        """Get the predefined configuration set.

        Returns:
            Dict[torch_utils.ProcessorType, "GPTQConfig"]: The configuration of GPTQ.
        """
        pre_defined_configs: Dict[torch_utils.ProcessorType, GPTQConfig] = {}
        pre_defined_configs[torch_utils.ProcessorType.Client] = cls(
            use_layer_wise=True
        )  # , model_path=self.model_path)
        pre_defined_configs[torch_utils.ProcessorType.Server] = cls()
        return pre_defined_configs


def get_default_gptq_config(processor_type: Optional[Union[str, torch_utils.ProcessorType]] = None) -> GPTQConfig:
    """Get the default configuration of GPTQ.

    Args:
        processor_type (Optional[Union[str, torch_utils.ProcessorType]], optional): The user-specified processor type.
            Defaults to None.

    Returns:
        GPTQConfig: GPTQConfig
    """
    process_type = torch_utils.get_processor_type_from_user_config(processor_type)
    return GPTQConfig.get_predefined_configs()[process_type]


######################## AWQ Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=AWQ, priority=PRIORITY_AWQ)
class AWQConfig(TorchBaseConfig):
    """Config class for AWQ.

    AWQ: Activation-aware Weight Quantization for LLM Compression and Acceleration.
    https://arxiv.org/abs/2306.00978
    """

    supported_configs: List[OperatorConfig] = []
    name = AWQ

    def __init__(
        self,
        dtype: str = "int",
        bits: int = 4,
        use_sym: bool = True,
        group_size: int = 32,
        group_dim: int = 1,
        use_full_range: bool = False,
        use_mse_search: bool = False,
        use_layer_wise: bool = False,
        model_path: str = "",
        # double quant
        use_double_quant: bool = False,
        double_quant_dtype: str = "int",
        double_quant_bits: int = 8,  # not available when double_quant_dtype is not 'int'
        double_quant_use_sym: bool = True,
        double_quant_group_size: int = 256,
        # quant lm_head
        quant_lm_head: bool = False,
        # awq
        use_auto_scale: bool = True,
        use_auto_clip: bool = True,
        folding: bool = False,
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        absorb_layer_dict: dict = {},
        **kwargs,
    ):
        """Init AWQ weight-only quantization config.

        Args:
            dtype (str): Data type for weights, default is "int".
            bits (int): Number of bits used to represent weights, default is 4.
            use_sym (bool): Indicates whether weights are symmetric, default is True.
            group_size (int): Size of weight groups, default is 32.
            group_dim (int): Dimension for grouping, default is 1.
            use_full_range (bool): Enables full range for activations, default is False.
            use_mse_search (bool): Enables mean squared error (MSE) search, default is False.
            use_layer_wise (bool): Enables quantize model per layer. Defaults to False.
            model_path (str): Model path that is used to load state_dict per layer.
            use_double_quant (bool): Enables double quantization, default is False.
            double_quant_dtype (str): Data type for double_quant scale, default is "int".
            double_quant_bits (int): Number of bits used to represent double_quant scale, default is 4.
            double_quant_use_sym (bool): Indicates whether double_quant scale are symmetric, default is True.
            double_quant_group_size (int): Size of double_quant groups, default is 32.
            quant_lm_head (bool): Indicates whether quantize the lm_head layer in transformer, default is False.
            use_auto_scale (bool): Enables best scales search based on activation distribution, default is True.
            use_auto_clip (bool):  Enables clip range search. Defaults to True.
            folding(bool): Allow insert mul before linear when the scale cannot be absorbed by last layer,
              default is False.
            absorb_layer_dict (dict): The layer dict that scale can be absorbed, default is {}.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
              Default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.dtype = dtype
        self.bits = bits
        self.use_sym = use_sym
        self.group_size = group_size
        self.group_dim = group_dim
        self.use_full_range = use_full_range
        self.use_mse_search = use_mse_search
        self.use_layer_wise = use_layer_wise
        self.model_path = model_path
        # double quant
        self.use_double_quant = use_double_quant
        self.double_quant_bits = double_quant_bits
        self.double_quant_dtype = double_quant_dtype
        self.double_quant_use_sym = double_quant_use_sym
        self.double_quant_group_size = double_quant_group_size
        self.quant_lm_head = quant_lm_head
        self.use_auto_scale = use_auto_scale
        self.use_auto_clip = use_auto_clip
        self.folding = folding
        self.absorb_layer_dict = absorb_layer_dict
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations for AWQ.

        Returns:
            List[OperatorConfig]: List of supported operator configurations.
        """
        supported_configs = []
        # TODO(Yi)
        linear_awq_config = AWQConfig()
        operators = list(WOQ_WHITE_LIST)
        supported_configs.append(OperatorConfig(config=linear_awq_config, operators=operators))
        cls.supported_configs = supported_configs

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        if not self.quant_lm_head:
            self.set_local(
                LM_HEAD_NAMES, AWQConfig(dtype="fp32", use_layer_wise=self.use_layer_wise, model_path=self.model_path)
            )
        config_mapping = super().to_config_mapping(config_list, model_info)
        return config_mapping

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        filter_result = []
        for op_name, module in model.named_modules():
            if isinstance(module, WOQ_WHITE_LIST):
                pair = (op_name, type(module).__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "AWQConfig", List["AWQConfig"]]:
        """Get the default configuration set for tuning.

        Returns:
            Union[None, "AWQConfig", List["AWQConfig"]]: The configuration set for tuning.
        """
        # TODO fwk owner needs to update it.
        return AWQConfig(bits=[4, 6])


def get_default_awq_config() -> AWQConfig:
    """Generate the default awq config.

    Returns:
        the default awq config.
    """
    return AWQConfig()


######################## TEQ Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=TEQ, priority=PRIORITY_TEQ)
class TEQConfig(TorchBaseConfig):
    """Config class for TEQ.

    TEQ: Activation-aware Weight Quantization for LLM Compression and Acceleration.
    https://arxiv.org/abs/2306.00978
    """

    supported_configs: List[OperatorConfig] = []
    name = TEQ

    def __init__(
        self,
        dtype: str = "int",
        bits: int = 4,
        use_sym: bool = True,
        group_size: int = 32,
        group_dim: int = 1,
        use_full_range: bool = False,
        use_mse_search: bool = False,
        use_layer_wise: bool = False,
        # double quant
        use_double_quant: bool = False,
        double_quant_dtype: str = "int",
        double_quant_bits: int = 8,  # not available when double_quant_dtype is not 'int'
        double_quant_use_sym: bool = True,
        double_quant_group_size: int = 256,
        # quant lm_head
        quant_lm_head: bool = False,
        # teq
        absorb_to_layer: dict = {},
        folding: bool = True,
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init TEQ weight-only quantization config.

        Args:
            dtype (str): Data type for weights, default is "int".
            bits (int): Number of bits used to represent weights, default is 4.
            use_sym (bool): Indicates whether weights are symmetric, default is True.
            group_size (int): Size of weight groups, default is 32.
            group_dim (int): Dimension for grouping, default is 1.
            use_full_range (bool): Enables full range for activations, default is False.
            use_mse_search (bool): Enables mean squared error (MSE) search, default is False.
            use_layer_wise (bool): Enables quantize model per layer. Defaults to False.
            use_double_quant (bool): Enables double quantization, default is False.
            double_quant_dtype (str): Data type for double_quant scale, default is "int".
            double_quant_bits (int): Number of bits used to represent double_quant scale, default is 4.
            double_quant_use_sym (bool): Indicates whether double_quant scale are symmetric, default is True.
            double_quant_group_size (int): Size of double_quant groups, default is 32.
            quant_lm_head (bool): Indicates whether quantize the lm_head layer in transformers。 Default is False.
            absorb_to_layer (dict): The layer dict that scale can be absorbed, default is {}.
            folding(bool): Allow insert mul before linear when the scale cannot be absorbed by last layer,
              default is False.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
              Default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.dtype = dtype
        self.bits = bits
        self.use_sym = use_sym
        self.group_size = group_size
        self.group_dim = group_dim
        self.use_full_range = use_full_range
        self.use_mse_search = use_mse_search
        self.use_layer_wise = use_layer_wise
        # double quant
        self.use_double_quant = use_double_quant
        self.double_quant_bits = double_quant_bits
        self.double_quant_dtype = double_quant_dtype
        self.double_quant_use_sym = double_quant_use_sym
        self.double_quant_group_size = double_quant_group_size
        self.quant_lm_head = quant_lm_head
        self.absorb_to_layer = absorb_to_layer
        self.folding = folding
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations for TEQ.

        Returns:
            List[OperatorConfig]: List of supported operator configurations.
        """
        supported_configs = []
        # TODO(Yi)
        linear_teq_config = TEQConfig()
        operators = list(WOQ_WHITE_LIST)
        supported_configs.append(OperatorConfig(config=linear_teq_config, operators=operators))
        cls.supported_configs = supported_configs

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        if not self.quant_lm_head:
            self.set_local(LM_HEAD_NAMES, TEQConfig(dtype="fp32"))
        config_mapping = super().to_config_mapping(config_list, model_info)
        return config_mapping

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        filter_result = []
        for op_name, module in model.named_modules():
            if isinstance(module, WOQ_WHITE_LIST):
                pair = (op_name, type(module).__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "TEQConfig", List["TEQConfig"]]:
        """Get the default configuration set for tuning.

        Returns:
            Union[None, "TEQConfig", List["TEQConfig"]]: The configuration set for tuning.
        """
        # TODO fwk owner needs to update it.
        return TEQConfig(bits=[4, 6])


def get_default_teq_config() -> TEQConfig:
    """Generate the default teq config.

    Returns:
        the default teq config.
    """
    return TEQConfig()


######################## AUTOROUND Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=AUTOROUND, priority=PRIORITY_AUTOROUND)
class AutoRoundConfig(TorchBaseConfig):
    """Config class for AUTOROUND.

    AUTOROUND: Optimize Weight Rounding via Signed Gradient Descent for the Quantization of LLMs.
    https://arxiv.org/abs/2309.05516
    code: https://github.com/intel/auto-round
    """

    supported_configs: List[OperatorConfig] = []
    name = AUTOROUND

    def __init__(
        self,
        bits: int = None,
        group_size: int = None,
        use_sym: bool = None,
        dtype: str = None,
        act_bits: int = None,
        act_group_size: int = None,
        act_sym: bool = None,
        act_dtype: str = None,
        act_dynamic: bool = None,
        super_bits: int = None,
        super_group_size: int = None,
        # AUTOROUND
        enable_full_range: bool = False,
        batch_size: int = 8,
        amp: bool = True,
        lr_scheduler=None,
        enable_quanted_input: bool = True,
        enable_minmax_tuning: bool = True,
        lr: float = None,
        minmax_lr: float = None,
        low_gpu_mem_usage: bool = False,
        iters: int = 200,
        seqlen: int = 2048,
        nsamples: int = 128,
        sampler: str = "rand",
        seed: int = 42,
        nblocks: int = 1,
        gradient_accumulate_steps: int = 1,
        not_use_best_mse: bool = False,
        dynamic_max_gap: int = -1,
        scale_dtype: str = "fp16",
        use_layer_wise: bool = False,
        to_quant_block_names: list = None,
        export_format: str = "auto_round",
        # v0.4
        enable_norm_bias_tuning: bool = False,
        enable_torch_compile: bool = False,
        # v0.7
        scheme: str | dict = "W4A16",
        device_map: str | int | torch.device | dict = 0,
        # mllm
        quant_nontext_module: bool = False,
        extra_data_dir: str = None,
        processor=None,
        image_processor=None,
        template=None,
        truncation: bool = False,
        quant_lm_head: bool = False,
        # v0.8
        enable_adam: bool = False,
        # v0.9: auto scheme parameters
        target_bits: float = None,
        options: Union[str, list[Union[str]], tuple[Union[str], ...]] = ("MXFP4", "MXFP8"),
        shared_layers: Optional[Iterable[Iterable[str]]] = None,
        ignore_scale_zp_bits: bool = False,
        auto_scheme_method: str = "default",
        auto_scheme_device_map: str = None,
        auto_scheme_batch_size: int = None,
        output_dir: str = "./temp_auto_round",
        # Tuning space
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init AUTOROUND weight-only quantization config.

        Args:
            dtype (str): Data type for weights, default is None.
            bits (int): Number of bits used to represent weights, default is None.
            use_sym (bool): Indicates whether weights are symmetric, default is None.
            group_size (int): Size of weight groups, default is None.
            act_bits (int): Number of bits for activation quantization. Default is None.
            act_group_size (int): Group size for activation quantization. Default is None.
            act_sym (bool): Whether to use symmetric activation quantization. Default is None.
            act_dynamic (bool): Whether to use dynamic activation quantization. Default is None.
            act_dtype (Optional[str]): Data type for activation quantization. Default is None.
            enable_full_range (bool): Whether to enable full range quantization (default is False).
            batch_size (int): Batch size for training (default is 8).
            amp (bool): Whether to use automatic mixed precision (default is True).
            lr_scheduler: The learning rate scheduler to be used.
            enable_quanted_input (bool): Whether to use quantized input data (default is True).
            enable_minmax_tuning (bool): Whether to enable min-max tuning (default is True).
            lr (float): The learning rate (default is 0.005).
            minmax_lr (float): The learning rate for min-max tuning (default is None).
            low_gpu_mem_usage (bool): Whether to use low GPU memory (default is False).
            iters (int): Number of iterations (default is 200).
            seqlen (int): Length of the sequence.
            nsamples (int): Number of samples (default is 512).
            sampler (str): The sampling method (default is "rand").
            seed (int): The random seed (default is 42).
            nblocks (int): Number of blocks (default is 1).
            gradient_accumulate_steps (int): Number of gradient accumulation steps (default is 1).
            not_use_best_mse (bool): Whether to use mean squared error (default is False).
            dynamic_max_gap (int): The dynamic maximum gap (default is -1).
            scale_dtype (str): The data type of quantization scale to be used (default is "float16"), different kernels
              have different choices.
            use_layer_wise (bool): Enables quantize model per layer. Defaults to False.
            to_quant_block_names (list): A list whose elements are list of block's layer names to be quantized.
            export_format (str, optional): The format used for exporting the quantized model. Defaults to "auto_round".
            enable_norm_bias_tuning (bool): Whether to enable fast norm/layer_bias tuning.
            enable_torch_compile (bool): Whether to enable torch compile to optimize quant_block/layer, torch>=2.6 True.
            quant_nontext_module (bool): Whether to quantize nontext module.
            extra_data_dir (str): The path for extra data such as images, audio or videos.
            processor (transformers.AutoProcessor): Any multi-modal model will require an object to encode or
              decode the data that groups several modalities (among text, vision and audio).
              This is handled by objects called processors, which group together two or more processing objects such
              as tokenizers (for the text modality), image processors (for vision) and feature extractors (for audio).
            image_processor (Processor): Image processor for special model like llava.
            template (Template): The template to specify process for different mllms.
            truncation (bool): Activates truncation to cut input sequences longer than `max_length` to `max_length`.
            device_map: The device to be used for tuning.
            scheme (str| dict | QuantizationScheme ): A preset scheme that defines the quantization configurations.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
            target_bits (float): The target bit width for quantization (default is None).
                options (Union[str, list[Union[str]], tuple[Union[str], ...]]): The options for mixed-precision quantization.
                shared_layers (Optional[Iterable[Iterable[str]]]): The shared layers for mixed-precision quantization.
                ignore_scale_zp_bits (bool): Whether to ignore scale and zero-point bits (default is False).
                auto_scheme_method (str): The method for automatic scheme selection (default is "default").
                auto_scheme_device_map (str): The device map for automatic scheme selection (default is None).
                auto_scheme_batch_size (int): The batch size for automatic scheme selection (default is 8).
            output_dir (str): The output directory for temporary files (default is "./temp_auto_round").
        """
        super().__init__(white_list=white_list)
        # these two params are lists but not tunable
        self.non_tunable_params.extend(["options", "shared_layers"])

        self.enable_full_range = enable_full_range
        self.batch_size = batch_size
        self.bits = bits
        self.group_size = group_size
        self.use_sym = use_sym
        self.dtype = dtype
        self.act_bits = act_bits
        self.act_group_size = act_group_size
        self.act_sym = act_sym
        self.act_dtype = act_dtype
        self.act_dynamic = act_dynamic
        self.super_bits = super_bits
        self.super_group_size = super_group_size
        self.amp = amp
        self.enable_adam = enable_adam
        self.lr_scheduler = lr_scheduler
        self.enable_quanted_input = enable_quanted_input
        self.enable_minmax_tuning = enable_minmax_tuning
        self.lr = lr
        self.minmax_lr = minmax_lr
        self.low_gpu_mem_usage = low_gpu_mem_usage
        self.iters = iters
        self.seqlen = seqlen
        self.nsamples = nsamples
        self.sampler = sampler
        self.seed = seed
        self.nblocks = nblocks
        self.gradient_accumulate_steps = gradient_accumulate_steps
        self.not_use_best_mse = not_use_best_mse
        self.dynamic_max_gap = dynamic_max_gap
        self.scale_dtype = scale_dtype
        self.use_layer_wise = use_layer_wise
        self.to_quant_block_names = to_quant_block_names
        self.export_format = export_format
        self.enable_norm_bias_tuning = enable_norm_bias_tuning
        self.enable_torch_compile = enable_torch_compile
        self.quant_nontext_module = quant_nontext_module
        self.extra_data_dir = extra_data_dir
        self.processor = processor
        self.image_processor = image_processor
        self.template = template
        self.truncation = truncation
        self.scheme = scheme
        self.device_map = device_map
        self.quant_lm_head = quant_lm_head
        self.target_bits = target_bits
        self.options = options
        self.shared_layers = shared_layers
        self.ignore_scale_zp_bits = ignore_scale_zp_bits
        self.auto_scheme_method = auto_scheme_method
        self.auto_scheme_device_map = auto_scheme_device_map
        self.auto_scheme_batch_size = auto_scheme_batch_size
        self.output_dir = output_dir
        # add kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations for AutoRound.

        Returns:
            List[OperatorConfig]: List of supported operator configurations.
        """
        supported_configs = []
        # TODO(Yi)
        linear_AUTOROUND_config = AutoRoundConfig()
        operators = [torch.nn.Linear, torch.nn.functional.linear]
        supported_configs.append(OperatorConfig(config=linear_AUTOROUND_config, operators=operators))
        cls.supported_configs = supported_configs

    def get_params_dict(self):
        """Get a dictionary containing the parameters and their values for the current instance.

        Returns:
            A dictionary containing the parameters and their values.
        """
        result = dict()
        for param, value in self.__dict__.items():
            # tokenizer: excluded to avoid duplicating large tokenizer objects in per-layer configs.
            # The tokenizer is attached at the model level to ensure a single shared instance.
            if param not in ["_global_config", "_local_config", "_white_list", "tokenizer"]:
                result[param] = value
        return result

    @staticmethod
    def get_model_info(model: torch.nn.Module):
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        return None

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        config_mapping = OrderedDict({self.name: self})
        return config_mapping

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "AutoRoundConfig", List["AutoRoundConfig"]]:
        """Get the default configuration set for tuning.

        Returns:
            Union[None, "AutoRoundConfig", List["AutoRoundConfig"]]: The configuration set for tuning.
        """
        # TODO fwk owner needs to update it.
        return AutoRoundConfig(bits=[4, 6])

    @classmethod
    def get_predefined_configs(cls) -> Dict[torch_utils.ProcessorType, "AutoRoundConfig"]:
        """Get the predefined configuration set.

        Returns:
            Dict[torch_utils.ProcessorType, "AutoRoundConfig"]: The configuration of AutoRound.
        """
        pre_defined_configs: Dict[torch_utils.ProcessorType, AutoRoundConfig] = {}
        pre_defined_configs[torch_utils.ProcessorType.Client] = cls(use_layer_wise=True)
        pre_defined_configs[torch_utils.ProcessorType.Server] = cls()
        return pre_defined_configs


def get_default_AutoRound_config(processor_type: Optional[Union[str, torch_utils.ProcessorType]] = None) -> RTNConfig:
    """Get the default configuration of AutoRound.

    Args:
        processor_type (Optional[Union[str, torch_utils.ProcessorType]], optional): The user-specified processor type.
            Defaults to None.

    Returns:
        AutoRoundConfig: AutoRoundConfig
    """
    process_type = torch_utils.get_processor_type_from_user_config(processor_type)
    return AutoRoundConfig.get_predefined_configs()[process_type]


######################## MX Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=MX_QUANT)
class MXQuantConfig(TorchBaseConfig):
    """Config class for MX quantization."""

    supported_configs: List[OperatorConfig] = []
    name = MX_QUANT

    def __init__(
        self,
        w_dtype: str = "int8",
        act_dtype: str = "int8",
        out_dtype: str = "bfloat16",
        blocksize: int = 32,
        round_method: str = "nearest",
        weight_only: bool = False,
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init MX quantization config.

        Args:
            w_dtype (str): Data type for weights, default is "int8".
            act_dtype (str): Data type for activations, default is "int8".
            out_dtype (str): Data type for outputs, default is "bfloat16".
            blocksize (int): Granularity to share the scale, default is 32.
            round_method (str): Round method, default is "nearest".
            weight_only (bool): Whether implement weight_only, default is False.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
              Default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.w_dtype = w_dtype
        self.act_dtype = act_dtype
        self.out_dtype = out_dtype
        self.blocksize = blocksize
        self.round_method = round_method
        self.weight_only = weight_only
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations."""
        supported_configs = []
        linear_mx_config = MXQuantConfig(
            w_dtype=[
                "int8",
                "int4",
                "int2",
                "fp8_e5m2",
                "fp8_e4m3",
                "fp6_e3m2",
                "fp6_e2m3",
                "fp4",
                "float16",
                "bfloat16",
                "float32",
            ],
            act_dtype=[
                "int8",
                "int4",
                "int2",
                "fp8_e5m2",
                "fp8_e4m3",
                "fp6_e3m2",
                "fp6_e2m3",
                "fp4",
                "float16",
                "bfloat16",
                "float32",
            ],
            out_dtype=["bfloat16", "float16", "float32"],
            blocksize=[2, 4, 8, 16, 32, 64, 128, 256, 512],
            round_method=["nearest", "dither", "floor", "even"],
            weight_only=[True, False],
        )
        operators = [torch.nn.Linear, torch.nn.functional.linear]
        supported_configs.append(OperatorConfig(config=linear_mx_config, operators=operators))
        cls.supported_configs = supported_configs

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        white_list = (
            torch.nn.Linear,
            torch.nn.functional.linear,
        )

        filter_result = []
        for op_name, module in model.named_modules():
            if module.__class__ in white_list:
                pair = (op_name, type(module).__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "MXQuantConfig", List["MXQuantConfig"]]:
        """Get the default configuration set for tuning."""
        return MXQuantConfig(weight_only=[False, True])


def get_default_mx_config() -> MXQuantConfig:
    """Generate the default mx config.

    Returns:
        the default rtn config.
    """
    return MXQuantConfig()


######################## Dynamic Quant Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=PT2E_DYNAMIC_QUANT)
class DynamicQuantConfig(TorchBaseConfig):
    """Config class for dynamic quantization."""

    name = PT2E_DYNAMIC_QUANT
    supported_configs: List[OperatorConfig] = []

    def __init__(
        self,
        w_dtype: str = "int8",
        w_sym: bool = True,
        w_granularity: str = "per_tensor",
        w_algo: str = "minmax",
        act_dtype: str = "uint8",
        act_sym: bool = False,
        act_granularity: str = "per_tensor",
        act_algo: str = "kl",
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init Dynamic Quant Configs."""
        super().__init__(white_list=white_list)
        self.w_dtype = w_dtype
        self.w_sym = w_sym
        self.w_granularity = w_granularity
        self.w_algo = w_algo
        self.act_dtype = act_dtype
        self.act_sym = act_sym
        self.act_granularity = act_granularity
        self.act_algo = act_algo
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations."""
        supported_configs = []
        linear_static_config = cls()
        operators = [torch.nn.Linear]
        supported_configs.append(OperatorConfig(config=linear_static_config, operators=operators))
        cls.supported_configs = supported_configs

    @staticmethod
    def get_model_info(model: torch.nn.Module, example_inputs=None):
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        return None

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        config_mapping = OrderedDict({self.name: self})
        return config_mapping

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "DynamicQuantConfig", List["DynamicQuantConfig"]]:
        """Get the default configuration set for tuning."""
        return cls(act_sym=[True, False], act_algo=["kl", "minmax"])


def get_default_dynamic_config() -> DynamicQuantConfig:
    """Generate the default dynamic quant config.

    Returns:
        the default dynamic quant config.
    """
    return DynamicQuantConfig()


######################## Static Quant Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=STATIC_QUANT)
class INT8StaticQuantConfig(TorchBaseConfig):
    """Config class for static quantization."""

    name = STATIC_QUANT
    supported_configs: List[OperatorConfig] = []

    def __init__(
        self,
        w_dtype: str = "int8",
        w_sym: bool = True,
        w_granularity: str = "per_channel",
        w_algo: str = "minmax",
        act_dtype: str = "uint8",
        act_sym: bool = False,
        act_granularity: str = "per_tensor",
        act_algo: str = "minmax",
        excluded_precisions: list = [],
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        model_info: Optional[List[Tuple[str, Callable]]] = None,
        **kwargs,
    ):
        """Init StaticQuant Config.

        Args:
            w_dtype (str): Data type for weights, default is "int8".
            w_sym (bool): Whether to use symmetric quantization for weights, default is True.
            w_granularity (str): Level of quantization granularity for weights, default is "per_channel".
            w_algo (str): Quatization algorithm used to compute parameters for weights, default is "minmax".
            act_dtype (str): Data type for activations, default is "uint8".
            act_sym (bool): Whether to use symmetric quantization for activations, default is False.
            act_granularity (str): Level of quantization granularity for activations, default is "per_channel".
            act_algo (str): Quatization algorithm used to compute parameters for activations, default is "minmax".
            excluded_precisions (list): Precisions to be excluded, Default value is empty list.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
                                                                 Default is DEFAULT_WHITE_LIST.
            model_info (Optional): used to keep model info for XPU device.  # TODO: should be removed from input arguments
        """
        super().__init__(white_list=white_list)
        self.w_dtype = w_dtype
        self.w_sym = w_sym
        self.w_granularity = w_granularity
        self.w_algo = w_algo
        self.act_dtype = act_dtype
        self.act_sym = act_sym
        self.act_granularity = act_granularity
        self.act_algo = act_algo
        self.excluded_precisions = excluded_precisions
        self.model_info = model_info
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations."""
        supported_configs = []
        linear_static_config = INT8StaticQuantConfig()
        operators = [torch.nn.Linear]
        supported_configs.append(OperatorConfig(config=linear_static_config, operators=operators))
        cls.supported_configs = supported_configs

    @staticmethod
    def get_model_info_for_ipex(model: torch.nn.Module, example_inputs) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        from neural_compressor.torch.algorithms.static_quant import get_quantizable_ops_recursively

        _, _, _, _, model_info = get_quantizable_ops_recursively(model, example_inputs=example_inputs)
        return model_info

    def get_model_info_for_ipex_xpu(self, model: torch.nn.Module) -> List[Tuple[str, Callable]]:  # pragma: no cover
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        if self.model_info:
            return self.model_info
        else:
            white_list = torch.quantization.quantization_mappings.get_default_qconfig_propagation_list()
            filter_result = []
            for op_name, module in model.named_modules():
                if type(module) in white_list:
                    pair = (op_name, type(module).__name__)
                    filter_result.append(pair)
            logger.debug(f"Get model info: {filter_result}")
            self.model_info = filter_result
            return filter_result

    def get_model_info(self, model: torch.nn.Module, example_inputs=None) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

        if is_ipex_imported():
            if auto_detect_accelerator().current_device() == "cpu":
                return INT8StaticQuantConfig.get_model_info_for_ipex(model, example_inputs)
            else:
                return INT8StaticQuantConfig.get_model_info_for_ipex_xpu(self, model)

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        if is_ipex_imported():
            return super().to_config_mapping(config_list, model_info)
        config_mapping = OrderedDict({self.name: self})
        return config_mapping

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "INT8StaticQuantConfig", List["INT8StaticQuantConfig"]]:
        """Get the default configuration set for tuning."""
        return INT8StaticQuantConfig(act_sym=[True, False], act_algo=["kl", "minmax"])


def get_default_static_config() -> INT8StaticQuantConfig:
    """Generate the default static quant config.

    Returns:
        the default static quant config.
    """
    return INT8StaticQuantConfig()


######################## Smooth Quant Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=SMOOTH_QUANT)
class SmoothQuantConfig(TorchBaseConfig):
    """Config class for smooth quantization."""

    name = SMOOTH_QUANT
    supported_configs: List[OperatorConfig] = []

    def __init__(
        self,
        w_dtype: str = "int8",
        w_sym: bool = True,
        w_granularity: str = "per_channel",
        w_algo: str = "minmax",
        act_dtype: str = "uint8",
        act_sym: bool = False,
        act_granularity: str = "per_tensor",
        act_algo: str = "minmax",
        excluded_precisions: list = [],
        alpha: float = 0.5,
        folding: bool = False,
        # below for autotune
        scale_sharing: bool = False,
        init_alpha: float = 0.5,
        alpha_min: float = 0.0,
        alpha_max: float = 1.0,
        alpha_step: float = 0.1,
        shared_criterion: str = "max",
        do_blockwise: bool = False,
        auto_alpha_args: dict = None,
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init SmoothQuant Config.

        Args:
            w_dtype (str): Data type for weights, default is "int8".
            w_sym (bool): Whether to use symmetric quantization for weights, default is True.
            w_granularity (str): Level of quantization granularity for weights, default is "per_channel".
            w_algo (str): Quatization algorithm used to compute parameters for weights, default is "minmax".
            act_dtype (str): Data type for activations, default is "uint8".
            act_sym (bool): Whether to use symmetric quantization for activations, default is False.
            act_granularity (str): Level of quantization granularity for activations, default is "per_channel".
            act_algo (str): Quatization algorithm used to compute parameters for activations, default is "minmax".
            excluded_precisions (list): Precisions to be excluded, Default value is empty list.
            alpha (float): Value to balance input and weight quantization error, between 0 and 1, default is 0.5.
            folding (bool): Whether to fold mul into the previous layer, default is False.
            scale_sharing (bool): Whether share the same scale for layers with the same input, default is False.
            init_alpha (float): Value to get baseline quantization error for auto-tuning, default is 0.5.
            alpha_min (float): Min value of auto-tuning alpha search space, default is 0.0.
            alpha_max (float): Max value of auto-tuning alpha search space, default is 1.0.
            alpha_step (float): Step_size of auto-tuning alpha search space, default is 0.1.
            shared_criterion (str): Criterion for input LayerNorm op of a transformer block, default is "max".
            do_blockwise (bool): Whether to enable block-wise auto-tuning, default is False.
            auto_alpha_args (bool): Arguments for auto alpha searching, default is None.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
                                                                 Default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.w_dtype = w_dtype
        self.w_sym = w_sym
        self.w_granularity = w_granularity
        self.w_algo = w_algo
        self.act_dtype = act_dtype
        self.act_sym = act_sym
        self.act_granularity = act_granularity
        self.act_algo = act_algo
        self.excluded_precisions = excluded_precisions
        self.alpha = alpha
        self.folding = folding
        # below for autotune
        self.scale_sharing = scale_sharing
        self.init_alpha = init_alpha
        self.alpha_min = alpha_min
        self.alpha_max = alpha_max
        self.alpha_step = alpha_step
        self.shared_criterion = shared_criterion
        self.do_blockwise = do_blockwise
        self.auto_alpha_args = {
            "init_alpha": self.init_alpha,
            "alpha_min": self.alpha_min,
            "alpha_max": self.alpha_max,
            "alpha_step": self.alpha_step,
            "shared_criterion": self.shared_criterion,
            "do_blockwise": self.do_blockwise,
        }
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations."""
        supported_configs = []
        # TODO(Yi)
        linear_sq_config = SmoothQuantConfig()
        operators = [torch.nn.Linear]
        supported_configs.append(OperatorConfig(config=linear_sq_config, operators=operators))
        cls.supported_configs = supported_configs

    def get_model_info(self, model: torch.nn.Module, example_inputs) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        from neural_compressor.torch.algorithms.smooth_quant import get_quantizable_ops_recursively

        model_info, cfgs, op_infos_from_cfgs, output_tensor_id_op_name = get_quantizable_ops_recursively(
            model, example_inputs, alpha=self.alpha, act_algo=self.act_algo, inplace=True
        )
        model.cfgs, model.op_infos_from_cfgs, model.output_tensor_id_op_name = (
            cfgs,
            op_infos_from_cfgs,
            output_tensor_id_op_name,
        )
        return model_info

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "SmoothQuantConfig", List["SmoothQuantConfig"]]:
        """Get the default configuration set for tuning."""
        import numpy as np

        return SmoothQuantConfig(
            alpha=np.arange(0.1, 1.0, 0.1).tolist(),
            folding=[True, False],
            scale_sharing=[True, False],
            excluded_precisions=[["bf16"]],
        )


def get_default_sq_config() -> SmoothQuantConfig:
    """Generate the default smoothquant config.

    Returns:
        the default smoothquant config.
    """
    return SmoothQuantConfig()


######################## HQQ Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=HQQ, priority=PRIORITY_HQQ)
class HQQConfig(TorchBaseConfig):
    """Configuration class for Half-Quadratic Quantization (HQQ).

    HQQ is a quantization algorithm that reduces the precision of weights and activations in neural networks.
    For more details, refer to the blog: https://mobiusml.github.io/hqq_blog/
    and the code: https://github.com/mobiusml/hqq
    """

    name = HQQ
    supported_configs: List[OperatorConfig] = []

    def __init__(
        self,
        dtype: str = "int",
        bits: int = 4,
        group_size: int = 64,
        quant_zero: bool = True,
        quant_scale: bool = False,
        scale_quant_group_size: int = 128,
        quant_lm_head: bool = False,
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Initialize HQQConfig.

        Args:
            dtype (str): Data type for quantization. Default is "int".
            bits (int): Number of bits for quantization. Default is 4.
            group_size (int): Group size for quantization. Default is 64.
            quant_zero (bool): Whether to quantize zero values. Default is True.
            quant_scale (bool): Whether to quantize scale values. Default is False.
            scale_quant_group_size (int): Group size for scale quantization. Default is 128.
            quant_lm_head (bool): Whether to quantize the language model head. Default is False.
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): White list of operator names or module types.
                Default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.dtype = dtype
        self.bits = bits
        self.group_size = group_size
        self.quant_zero = quant_zero
        self.quant_scale = quant_scale
        self.scale_quant_group_size = scale_quant_group_size
        self.quant_lm_head = quant_lm_head
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configurations for HQQ.

        Returns:
            List[OperatorConfig]: List of supported operator configurations.
        """
        supported_configs = []
        linear_hqq_config = HQQConfig()
        operators = list(WOQ_WHITE_LIST)
        supported_configs.append(OperatorConfig(config=linear_hqq_config, operators=operators))
        cls.supported_configs = supported_configs

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        filter_result = []
        for op_name, module in model.named_modules():
            if isinstance(module, WOQ_WHITE_LIST):
                pair = (op_name, type(module).__name__)
                filter_result.append(pair)
        return filter_result

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]:
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        if not self.quant_lm_head:
            self.set_local(LM_HEAD_NAMES, HQQConfig(dtype="fp32"))
        config_mapping = super().to_config_mapping(config_list, model_info)
        return config_mapping

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "HQQConfig", List["HQQConfig"]]:
        """Get the configuration set for tuning.

        Returns:
            Union[None, "HQQConfig", List["HQQConfig"]]: The configuration set for tuning.
        """
        return HQQConfig(bits=[4, 8])


def get_default_hqq_config() -> HQQConfig:
    """Generate the default HQQ config.

    Returns:
        the default HQQ config.
    """
    return HQQConfig()


######################## FP8 Quant Config ###############################
from neural_compressor.torch.algorithms.fp8_quant._core.patching_common import get_white_list


@register_config(framework_name=FRAMEWORK_NAME, algo_name=FP8_QUANT)
class FP8Config(TorchBaseConfig):
    """Config class for FP8 quantization."""

    name = FP8_QUANT

    def __init__(
        self,
        dump_stats_path: str = "./hqt_output/measure",
        fp8_config: str = "E4M3",
        hp_dtype: str = "bf16",
        blocklist: dict = {"names": [], "types": ()},
        allowlist: dict = {"names": [], "types": get_white_list()},
        mode: str = "AUTO",
        scale_method: Union[str, dict] = "maxabs_hw",
        scale_params: dict = {},
        observer: str = "maxabs",
        mod_dict: dict = {},
        measure_exclude: str = "OUTPUT",
        fake_quant: bool = False,
        use_qdq: bool = False,
        scale_format: str = "scalar",
        measure_on_hpu: bool = True,
        **kwargs,
    ):
        """Initializing FP8Config.

        Args:
            dump_stats_path (str, optional): The file folder and file prefix to save measurement info. Defaults to "./hqt_output/measure".
            fp8_config (str, optional): The data type of fp8. Defaults to "E4M3".
            hp_dtype (str, optional): The high precision data type used in fp8 quantization. Defaults to "bf16".
            blocklist (dict, optional): Whether to skip fp8 quantization for specific op names or types, name could be substring. Defaults to {"names": [], "types": ()}.
            allowlist (dict, optional): Whether to execute fp8 quantization for specific op names or types. Defaults to {"names": [], "types": FP8_WHITE_LIST}.
            mode (str, optional): Choose the quantization mode. Defaults to "AUTO".
            scale_method (str or dict, optional): Select method used to generate scale from calibration info. Can be a string or a dict. Defaults to "maxabs_hw".
            scale_params (dict, optional): Scaling parameters that override the default ones for specific modules. Defaults to {}.
            observer (str, optional): Params of scales. Defaults to "maxabs".
            mod_dict (dict, optional): The dict of modules to quantize. Defaults to {}.
            measure_exclude (str, optional): Select INPUT/OUTPUT to be exculded by measurement. Defaults to "OUTPUT".
            fake_quant (bool, optional): whether to execute fake quantization, a little bit different with use_qdq, used for training. Defaults to False.
            use_qdq (bool, optional): whether to execute Q/DQ quantization. Defaults to False.
        """
        super().__init__()
        self.dump_stats_path = dump_stats_path
        self.fp8_config = fp8_config
        self.hp_dtype = hp_dtype
        self.blocklist = blocklist
        self.allowlist = allowlist
        self.mode = mode
        self.scale_method = scale_method
        self.scale_params = scale_params
        self.observer = observer
        self.mod_dict = mod_dict
        self._json_file = None
        self.measure_exclude = measure_exclude
        self.fake_quant = str(fake_quant)
        self.use_qdq = str(use_qdq)
        self.scale_format = scale_format
        self.measure_on_hpu = measure_on_hpu
        # add kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def measure(self):
        """Check whether the mode is for measurement."""
        return self.mode == "MEASURE"

    @property
    def quantize(self):
        """Check whether the mode is for quantization."""
        return self.mode == "QUANTIZE"

    @property
    def json_file(self):
        """Get the path of json file."""
        return self._json_file

    @json_file.setter
    def json_file(self, json_file):
        """Set the path of json file."""
        self._json_file = json_file

    @classmethod
    def from_json_file(cls, filename):
        """Set configuration from json file."""
        with open(filename, "r", encoding="utf-8") as file:
            config_dict = json.load(file)
        config = cls.from_dict(config_dict)
        config.json_file = filename
        return config

    def save_temp_json_file(self):
        """Save configuration to a temporary json file."""
        import tempfile
        from pathlib import Path

        json_file_tmp = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
        self.to_json_file(json_file_tmp.name)
        self._json_file = json_file_tmp.name

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "FP8Config", List["FP8Config"]]:
        """Get the configuration set for tuning."""
        # just a simple example here
        # usually write parameter combinations that are more suitable to tune based on experience.
        return FP8Config()

    @classmethod
    def register_supported_configs(cls) -> List:
        pass

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        filter_result = []
        for op_name, module in model.named_modules():
            if (
                module.__class__.__name__ in get_white_list()
                or module.__class__.__name__.split("Patched")[-1] in get_white_list()
            ):
                pair = (op_name, module.__class__.__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    def to_config_mapping(self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None):
        """Convert the configuration to a mapping.

        Args:
            config_list (List[BaseConfig]): List of base configurations. Default is None.
            model_info (List[Tuple[str, str]]): List of tuples containing the name and type of each module in the model.
                Default is None.

        Returns:
            OrderedDictType[Union[str, str], OrderedDictType[str, BaseConfig]]: The configuration mapping.
        """
        if self.json_file is None:
            self.save_temp_json_file()
        config_mapping = OrderedDict()
        if config_list is None:
            config_list = [self]
        for config in config_list:
            for op_name, op_type in model_info:
                config_mapping[(op_name, op_type)] = self
        return config_mapping


def get_default_fp8_config() -> FP8Config:
    """Generate the default fp8 config.

    Returns:
        the default fp8 config.
    """
    return FP8Config()


def get_default_fp8_config_set() -> FP8Config:
    """Generate the default fp8 config set.

    Returns:
        the default fp8 config.
    """
    return FP8Config.get_config_set_for_tuning()


######################## Hybrid GPTQ Quant Config ###############################


@register_config(framework_name=FRAMEWORK_NAME, algo_name=HYBRID_GPTQ)
class HybridGPTQConfig(FP8Config):
    """Config class for Hybrid Precision GPTQ quantization.

    Currently supports running 4bit weights which have been quantized by GPTQ, during which
    the weights have been double quantized from high precision, to fp8, to int4.
    The activations will be quantized to fp8.
    """

    name = HYBRID_GPTQ

    def __init__(
        self,
        **kwargs,
    ):
        """Initializing Hybrid GPTQ Config."""
        super().__init__(**kwargs)

    @staticmethod
    def convert_from_fp8(config):
        new_self = HybridGPTQConfig()
        for attr, value in vars(config).items():
            setattr(new_self, attr, value)
        return new_self


######################## MixedPrecision Config ###############################
@register_config(framework_name=FRAMEWORK_NAME, algo_name=MIXED_PRECISION)
class MixedPrecisionConfig(TorchBaseConfig):
    """Config class for mixed-precision."""

    name = MIXED_PRECISION
    supported_configs: List[OperatorConfig] = []
    supported_half_precision_ops = (
        torch.nn.Linear,
        torch.nn.Conv1d,
        torch.nn.Conv2d,
        torch.nn.Conv3d,
    )

    def __init__(
        self,
        dtype: Union[str, List[str]] = "fp16",
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
        **kwargs,
    ):
        """Init MixedPrecision config.

        Args:
            dtype (str or list): The data type of mixed precision, default is fp16.
            white_list (list): White list of operator names or module types, default is DEFAULT_WHITE_LIST.
        """
        super().__init__(white_list=white_list)
        self.dtype = dtype
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configs."""
        supported_configs = []
        mixed_precision_config = MixedPrecisionConfig(
            dtype=["fp16", "bf16", "fp32"],
        )
        operators = cls.supported_half_precision_ops
        supported_configs.append(OperatorConfig(config=mixed_precision_config, operators=operators))
        cls.supported_configs = supported_configs

    @staticmethod
    def get_model_info(model: torch.nn.Module) -> List[Tuple[str, Callable]]:
        """Get information about the model.

        Args:
            model (torch.nn.Module): The model.

        Returns:
            List[Tuple[str, Callable]]: List of tuples containing the name and type of each module in the model.
        """
        white_list = tuple(MixedPrecisionConfig.supported_half_precision_ops)
        filter_result = []
        for op_name, module in model.named_modules():
            if isinstance(module, white_list):
                pair = (op_name, type(module).__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "MixedPrecisionConfig", List["MixedPrecisionConfig"]]:
        """Get a default config set for tuning."""
        return MixedPrecisionConfig(dtype=["fp16", "bf16", "fp32"])


def get_default_mixed_precision_config() -> MixedPrecisionConfig:
    """Generate the default mixed-precision config.

    Returns:
        the default mixed-precision config.
    """
    return MixedPrecisionConfig()


def get_default_mixed_precision_config_set() -> MixedPrecisionConfig:
    """Generate the default mixed-precision config set.

    Returns:
        the default mixed-precision config.
    """
    return MixedPrecisionConfig.get_config_set_for_tuning()


##################### Algo Configs End ###################################


register_supported_configs_for_fwk(fwk_name=FRAMEWORK_NAME)


def get_all_registered_configs() -> Dict[str, BaseConfig]:
    """Get all registered configs."""
    registered_configs = config_registry.get_all_configs()
    return registered_configs.get(FRAMEWORK_NAME, {})


# =============================================================================
# Tuning Config
# =============================================================================


######################## WOQ Tuning Config ###############################
def get_woq_tuning_config() -> list:
    """Generate the config set for WOQ tuning.

    Returns:
        the list of WOQ quant config.
    """
    RTN_G32ASYM = RTNConfig(use_sym=False, group_size=32)
    AUTO_ROUND_CONFIG = AutoRoundConfig(use_sym=False, group_size=32, seqlen=512)
    GPTQ_G32ASYM = GPTQConfig(use_sym=False, group_size=32)
    AWQ_G32ASYM = AWQConfig(use_sym=False, group_size=32)
    return [RTN_G32ASYM, AUTO_ROUND_CONFIG, GPTQ_G32ASYM, AWQ_G32ASYM]


CONFIGS_FOR_STATIC_QUANT_MAPPING = OrderedDict(
    [
        # Configs for static quant mapping
        (STATIC_QUANT, INT8StaticQuantConfig),
        (FP8_QUANT, FP8Config),
    ]
)


class StaticQuantConfig(TorchBaseConfig):
    _model_mapping = CONFIGS_FOR_STATIC_QUANT_MAPPING

    def __new__(
        self,
        *args,
        **kwargs,
    ):
        dtype = kwargs.get("fp8_config", None)
        if dtype is not None:
            config_cls = self._model_mapping[FP8_QUANT]
        else:
            config_cls = self._model_mapping[STATIC_QUANT]
        return config_cls(*args, **kwargs)

    @classmethod
    def get_config_set_for_tuning(cls, dtype="int8"):
        """Map to different config set for tuning."""
        # dtype = "fp8", "int8"
        if dtype == "fp8":
            return cls._model_mapping[FP8_QUANT].get_config_set_for_tuning()
        elif dtype == "int8":
            return cls._model_mapping[STATIC_QUANT].get_config_set_for_tuning()
        else:
            raise ValueError(f"Unsupported dtype: {dtype}, allowed values are 'fp8' and 'int8'.")


# TODO: support more mappings configurations.
# Default map for swapping float module to qat modules
DEFAULT_QAT_MODULE_MAPPINGS: dict[Callable, Any] = {
    torch.nn.Linear: "MXFP8",
}

QAT_MODULE_MAPPINGS: dict[Callable, Any] = {
    torch.nn.Linear: ["MXFP8", "MXFP4"],
}


def get_default_qat_module_mappings() -> dict[Callable, Any]:
    """Get default module mapping for quantization aware training."""
    return copy.deepcopy(DEFAULT_QAT_MODULE_MAPPINGS)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Export model for quantization."""

from functools import partial
from typing import Any, Dict, Optional, Tuple, Union

import torch
from torch.fx.graph_module import GraphModule

from neural_compressor.common.utils import logger
from neural_compressor.torch.utils import (
    TORCH_VERSION_2_2_2,
    TORCH_VERSION_2_7_0,
    TORCH_VERSION_2_8_0,
    get_torch_version,
    is_ipex_imported,
)

__all__ = ["export", "export_model_for_pt2e_quant"]


def export_model_for_pt2e_quant(
    model: torch.nn.Module,
    example_inputs: Tuple[Any],
    dynamic_shapes: Optional[Union[Dict[str, Any], Tuple[Any]]] = None,
) -> Optional[GraphModule]:
    """Exports a eager model for PT2E quantization.

    Args:
        model (torch.nn.Module): The PyTorch model to be exported.
        example_inputs (Tuple[Any]): Example inputs to the model.
        dynamic_shapes (Optional[Union[Dict[str, Any], Tuple[Any]]], optional):
            Dynamic shapes for the model inputs. Defaults to None.

    Returns:
        Optional[GraphModule]: The exported model as a GraphModule.

    Raises:
        AssertionError: If `example_inputs` is not a tuple.
    """
    assert isinstance(example_inputs, tuple), f"Expected `example_inputs` to be a tuple, got {type(example_inputs)}"
    # Set the model to eval mode
    model = model.eval()
    exported_model = None
    try:
        with torch.no_grad():
            # Note 1: `capture_pre_autograd_graph` is also a short-term API, it will be
            # updated to use the official `torch.export` API when that is ready.
            cur_version = get_torch_version()
            if cur_version >= TORCH_VERSION_2_8_0:
                export_func = torch.export.export
                export_func = partial(export_func, strict=True)
            elif cur_version >= TORCH_VERSION_2_7_0:
                export_func = torch.export.export_for_training
            else:
                export_func = torch._export.capture_pre_autograd_graph
            if cur_version <= TORCH_VERSION_2_2_2:  # pragma: no cover
                logger.warning(
                    (
                        "`dynamic_shapes` is not supported in the current version(%s) of PyTorch,"
                        "If you want to use `dynamic_shapes` to export model, "
                        "please upgrade to 2.3.0 or later."
                    ),
                    cur_version,
                )
                exported_model = export_func(model, args=example_inputs)
            else:
                exported_model = export_func(  # pylint: disable=E1123
                    model, args=example_inputs, dynamic_shapes=dynamic_shapes
                )
            if cur_version >= TORCH_VERSION_2_7_0:
                exported_model = exported_model.module()
            exported_model._exported = True
            logger.info("Exported the model to Aten IR successfully.")
    except Exception as e:
        logger.error(f"Failed to export the model: {e}")

    return exported_model


def export(
    model: torch.nn.Module,
    example_inputs: Tuple[Any],
    dynamic_shapes: Optional[Union[Dict[str, Any], Tuple[Any]]] = None,
) -> Optional[GraphModule]:
    """Unified export function for quantization.

    Args:
        model (torch.nn.Module): The model to be exported.
        example_inputs (Tuple[Any]): Example inputs to the model.
        dynamic_shapes (Optional[Union[Dict[str, Any], Tuple[Any]]], optional):
            Dynamic shapes for the model. Defaults to None.

    Returns:
        Optional[GraphModule]: The exported model for quantization.
    """
    if not is_ipex_imported():
        model = export_model_for_pt2e_quant(model, example_inputs, dynamic_shapes)
        model.dynamic_shapes = dynamic_shapes
        return model
    else:
        # TODO, add `export` for ipex
        pass



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util
import os
from ..model_configs import ModuleInfo, ModuleType
from .._quant_common.helper_modules import *
from ..utils.logger import logger

from neural_compressor.torch.algorithms.fp8_quant.model_configs import (
    get_patched_module_table,
    get_patched_module_type_table,
)
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator, INCAcceleratorType
deepspeed_exists = False
if importlib.util.find_spec("deepspeed"):  # check if deepspeed is installed
    deepspeed_exists = True

mod_default_dict = get_patched_module_table()
mod_types = get_patched_module_type_table()


def get_white_list():
    return list(mod_default_dict.keys())


class ModInstInfo:
    def __init__(self, name, parent):
        self.name = name
        self.parent = parent


parent_child_mod_dict = {}


def generate_model_info(model):
    def create_mod_info_recursion(parent):
        for name, mod in parent.named_children():
            parent_child_mod_dict[mod] = ModInstInfo(name, parent)
            create_mod_info_recursion(mod)

    create_mod_info_recursion(model)


_mod_types = {
    "linear": ModuleType(1, ["weight"], 1, False),
    "row_parallel_linear": ModuleType(1, ["weight"], 2, True),
    "matmul": ModuleType(2, [], 1, False),
    "kv_cache": ModuleType(1, [], 1, False),
    "softmax": ModuleType(1, [], 1, True),
    "fused_sdpa": ModuleType(3, [], 2, True),
    "dynamic_moe": ModuleType(1, [], 1 + 8, True),
    "embedding": ModuleType(1, ["weight"], 1, False),
}


_mod_default_dict = {
    "Matmul": ModuleInfo("matmul", PatchedMatmul),
    "Linear": ModuleInfo("linear", PatchedLinear, supports_dynamic_quantization=True),
    "ParallelLMHead": ModuleInfo("linear", PatchedParallelLMHead, supports_dynamic_quantization=True),
    "RowParallelLinear": ModuleInfo("row_parallel_linear", PatchedRowParallelLinear, supports_dynamic_quantization=True),
    "ColumnParallelLinear": ModuleInfo("linear", PatchedColumnParallelLinear, supports_dynamic_quantization=True),
    "MergedColumnParallelLinear": ModuleInfo("linear", PatchedColumnParallelLinear, supports_dynamic_quantization=True),
    "QKVParallelLinear": ModuleInfo("linear", PatchedColumnParallelLinear, supports_dynamic_quantization=True),
    "FalconLinear": ModuleInfo("linear", PatchedLinear, supports_dynamic_quantization=True),
    "KVCache": ModuleInfo("kv_cache", PatchedKVCache),
    "VLLMKVCache": ModuleInfo("kv_cache", PatchedVLLMKVCache),
    "Conv2d": ModuleInfo("linear", PatchedConv2d),
    "LoRACompatibleLinear": ModuleInfo("linear", PatchedLoRACompatibleLinear, supports_dynamic_quantization=True),
    "LoRACompatibleConv": ModuleInfo("linear", PatchedLoRACompatibleConv),
    "Softmax": ModuleInfo("softmax", PatchedSoftmax),
    "BlockSoftmaxConstMax": ModuleInfo("softmax", PatchedBlockSoftmaxConstMax),
    "ModuleFusedSDPA": ModuleInfo("fused_sdpa", PatchedModuleFusedSDPA),
    "MoeMatmul": ModuleInfo("linear", PatchedMoeMatmul, supports_dynamic_quantization=True),
    "MoeFP8Matmul": ModuleInfo("linear", PatchedMoeFP8Matmul, supports_dynamic_quantization=True),
    "ReplicatedLinear": ModuleInfo("linear", PatchedReplicatedLinear, supports_dynamic_quantization=True),
    # Note: `no_quantize_op` indicates that this module is patched but does not require measurement or quantization.
    "FusedMoE": ModuleInfo("no_quantize_op", PatchedMixtralMoE, False, supports_dynamic_quantization=True),
    "SharedFusedMoE": ModuleInfo("no_quantize_op", PatchedMixtralMoE, False, supports_dynamic_quantization=True),
    "GaudiMixtralSparseMoeBlock": ModuleInfo("dynamic_moe", PatchedGaudiMixtralSparseMoeBlock),
    "GaudiDeepseekV3MoE": ModuleInfo("dynamic_moe", PatchedGaudiDeepseekV3MoE),
    "GaudiFP8Linear": ModuleInfo("linear", PatchedMoeFP8Matmul),
    "VllmMixtureOfExpertsOp": ModuleInfo("dynamic_moe", PatchedVllmMixtureOfExpertsOp, supports_dynamic_quantization=True),
    "VllmMixtureOfExpertsOpFP8": ModuleInfo("dynamic_moe", PatchedVllmMixtureOfExpertsOpFP8, supports_dynamic_quantization=True),
    "VllmMixtureOfExpertsOpFP8PerChannel": ModuleInfo("dynamic_moe", PatchedVllmMixtureOfExpertsOpFP8, supports_dynamic_quantization=True),
}


if deepspeed_exists:
    _mod_default_dict.update(
        {
            "LinearLayer": ModuleInfo("linear", PatchedLinear, supports_dynamic_quantization=True),
            "LinearAllreduce": ModuleInfo("linear", PatchedLinearAllReduce, supports_dynamic_quantization=True),
            "ScopedLinearAllReduce": ModuleInfo("linear", PatchedLinearAllReduce, supports_dynamic_quantization=True),
            "LmHeadLinearAllreduce": ModuleInfo("linear", PatchedLmHeadLinearAllreduce, supports_dynamic_quantization=True),
        }
    )


@functools.lru_cache(maxsize=None)
def _import_hpu_modules():
    from neural_compressor.torch.algorithms.fp8_quant.patched_module_base import (
        PATCHED_MODULE_TABLE, PATCHED_MODULE_TYPES_TABLE
    )
    cur_accelerator = auto_detect_accelerator()
    if not cur_accelerator.current_device_name().startswith("hpu"):
        return
    PATCHED_MODULE_TABLE["hpu"].update(_mod_default_dict)
    PATCHED_MODULE_TYPES_TABLE["hpu"].update(_mod_types)


@functools.lru_cache(maxsize=None)
def _import_xpu_modules():
    from neural_compressor.torch.algorithms.fp8_quant.patched_module_base import (
        PATCHED_MODULE_TABLE, PATCHED_MODULE_TYPES_TABLE
    )
    cur_accelerator = auto_detect_accelerator()
    if not cur_accelerator.current_device_name().startswith("xpu"):
        return
    PATCHED_MODULE_TABLE["xpu"].update({"Linear": ModuleInfo("linear", PatchedLinear),
                                        "Matmul": ModuleInfo("matmul", PatchedMatmul), })
    PATCHED_MODULE_TYPES_TABLE["xpu"].update({"linear": _mod_types["linear"]})


@functools.lru_cache(maxsize=None)
def _import_cpu_modules():
    from neural_compressor.torch.algorithms.fp8_quant.patched_module_base import (
        PATCHED_MODULE_TABLE, PATCHED_MODULE_TYPES_TABLE
    )
    cur_accelerator = auto_detect_accelerator()
    if not cur_accelerator.current_device_name().startswith("cpu"):
        return
    PATCHED_MODULE_TABLE["cpu"].update({"Linear": ModuleInfo("linear", PatchedLinear),
                                        "Conv2d": ModuleInfo("linear", PatchedConv2d),
                                        "EmbeddingBag": ModuleInfo("embedding", PatchedEmbeddingBag),
                                        })
    PATCHED_MODULE_TYPES_TABLE["cpu"].update({"linear": _mod_types["linear"], "embedding": _mod_types["embedding"]})


@functools.lru_cache(maxsize=None)
def _import_device_modules():
    cur_accelerator_type = auto_detect_accelerator().get_inc_accelerator_type()
    if cur_accelerator_type.value > INCAcceleratorType.GAUDI_MIN.value:
        _import_hpu_modules()
    elif cur_accelerator_type == INCAcceleratorType.XPU:
        _import_xpu_modules()
    elif cur_accelerator_type == INCAcceleratorType.CPU:
        _import_cpu_modules()
    else:
        logger.warning("No HPU or XPU devices were detected. No Patched Modules available.")


_import_device_modules()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 MIT HAN Lab
# This source code is licensed under the MIT license
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RTN quantization."""

import copy
from collections import OrderedDict

import torch

from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.utils import (
    get_accelerator,
    get_attr,
    get_model_device,
    is_transformers_imported,
    logger,
    set_attr,
    set_module,
)

from .modules import INCWeightOnlyLinear
from .utility import cast_fp8, quant_tensor, search_clip

if is_transformers_imported():
    import transformers


class RTNQuantizer(Quantizer):
    """RTN Quantizer."""

    def __init__(self, quant_config: OrderedDict = {}):
        """Init a RTNQuantizer object.

        Args:
            quant_config (OrderedDict, optional): quantization config for ops. Defaults to {}.
        """
        super().__init__(quant_config)

    @torch.no_grad()
    def prepare(self, model, *args, **kwargs):
        """Prepares a given model for quantization.

        Will return model directly in RTN algorithm.

        Args:
            model (torch.nn.Module): The model to be prepared.
        """
        return model

    @torch.no_grad()
    def convert(
        self,
        model,
        dtype="int",
        bits=4,
        scheme="sym",
        group_size=32,
        group_dim=1,
        quantile=1.0,
        use_full_range=False,
        use_mse_search=False,
        use_layer_wise=False,
        model_path="",
        quant_lm_head=False,
        *args,
        **kwargs,
    ):
        """Quant the model with round to nearest method and inplace is True.

        Args:
            model: torch module
            dtype (str, optional): select from int, nf4, fp4. Defaults to int.
            bits: num bits. Defaults to 4.
            scheme (str, optional): sym or asym. Defaults to "sym".
            group_size (int, optional): how many elements share one scale/zp. Defaults to 32.
            group_dim (int, optional):  0 means splitting output channel,
                                        1 means splitting input channel. Defaults to 1.
            quantile (float, optional): percentile of clip. Defaults to 1.0.
            use_full_range (bool, optional): Choose sym range whether use -2**(bits-1).
                                        Defaults to False.
            use_mse_search (bool, optional):  Whether to search clip range.
                                        Defaults to True.
            quant_lm_head (bool, optional):  Whether to quantize the lm_head layer.
                                        Defaults to False.

        Returns:
            model: fake quantized torch module
        """
        weight_config = self.quant_config
        device = get_accelerator(kwargs.pop("device", "auto")).current_device_name()
        model_device = get_model_device(model)  # return model on the same device

        # for transformers model. If lm_head is tied from embedding, we deepcopy it.
        if quant_lm_head and getattr(getattr(model, "config", None), "tie_word_embeddings", False):
            for key in model._tied_weights_keys:
                weight = get_attr(model, key)
                set_attr(model, key, copy.deepcopy(weight))

        assert isinstance(model, torch.nn.Module), "only support torch module"
        if is_transformers_imported():
            supported_layers = (torch.nn.Linear, transformers.Conv1D)
        else:
            supported_layers = (torch.nn.Linear,)
        # initialize global configuration
        double_quant_config = {
            "double_quant": kwargs.get("use_double_quant", False),
            "double_quant_dtype": kwargs.get("double_quant_dtype", "int"),
            "double_quant_bits": kwargs.get("double_quant_bits", 8),
            "double_quant_scheme": kwargs.get("double_quant_scheme", "sym"),
            "double_quant_group_size": kwargs.get("double_quant_group_size", 256),
        }
        use_optimum_format = kwargs.get("use_optimum_format", True)

        if use_layer_wise:
            from neural_compressor.common.utils import DEFAULT_WORKSPACE
            from neural_compressor.torch.algorithms.layer_wise.utils import get_path, load_module

            if model_path == "":
                model_path = model.path
            assert model_path, "model_path should not be None."
            model_path = get_path(model_path)

        for name, m in model.named_modules():
            if use_layer_wise and len(list(m.named_children())) == 0:
                load_module(model, name, model_path, device=device)
            if not isinstance(m, supported_layers):
                continue
            if name in weight_config:  # pragma: no cover
                # initialize op configuration
                dtype = weight_config[name].get("dtype", "int")
                if dtype == "fp32":
                    continue
                # Move modules to the accelerator device layer-by-layer
                if not use_layer_wise:
                    m.to(device)
                ### FP8 cast part
                if dtype in ["fp8_e5m2", "fp8_e5m2fnuz", "fp8_e4m3fn", "fp8_e4m3fnuz"]:
                    logger.debug("Cast module {} to FP8 using qdq mode, no scaling".format(name))
                    m.weight = cast_fp8(m.weight, dtype, use_qdq=True)
                    continue
                ####
                logger.debug("Apply RTN on module %s.", name)
                bits = weight_config[name].get("bits", 4)
                group_size = weight_config[name]["group_size"]
                scheme = weight_config[name]["scheme"]
                quantile = weight_config[name].get("quantile", 1.0)
                group_dim = weight_config[name]["group_dim"]
                use_full_range = weight_config[name]["use_full_range"]
                use_mse_search = weight_config[name]["use_mse_search"]
                use_optimum_format = kwargs.get("use_optimum_format", True)
                # double quant config
                double_quant_config = {
                    "double_quant": weight_config[name]["use_double_quant"],
                    "double_quant_dtype": weight_config[name]["double_quant_dtype"],
                    "double_quant_bits": weight_config[name]["double_quant_bits"],
                    "double_quant_scheme": weight_config[name]["double_quant_scheme"],
                    "double_quant_group_size": weight_config[name]["double_quant_group_size"],
                }
                if dtype != "int" and "int" in dtype:
                    bits = int(dtype.lstrip("int"))
                    dtype = "int"
            else:
                continue
            log_msg = (
                f"RTN quantization config: bits={bits}, group_size={group_size}, "
                + f"scheme={scheme}, quantile={quantile}"
            )
            if dtype != "int":
                log_msg += f", dtype={dtype}"
            elif scheme == "sym":  # nf4/fp4 is always [-7,7]
                log_msg += f", use_full_range={use_full_range}"
            if dtype == "fp32":
                continue
            logger.debug(f"RTN quantized module:{name, m}")
            logger.debug(log_msg)

            # for only group_dim is 0 or only `transformers.Conv1D`, we need transpose weight.
            if is_transformers_imported():
                transpose = (group_dim == 0) ^ (isinstance(m, transformers.Conv1D))
            else:
                transpose = group_dim == 0
            if transpose:
                weight = m.weight.detach().T.contiguous()
            else:
                weight = m.weight.detach()
            if use_mse_search:
                quantile = search_clip(m, bits, group_size, scheme, dtype, use_full_range)
            int_weight, scale, zp = quant_tensor(
                weight,
                dtype=dtype,
                bits=bits,
                group_size=group_size,
                scheme=scheme,
                quantile=quantile,
                return_int=True,
                full_range=use_full_range,
                **double_quant_config,
            )
            int_weight = int_weight.t_().contiguous() if transpose else int_weight
            scale = scale.t_().contiguous() if transpose else scale
            zp = zp.t_().contiguous() if transpose and zp is not None else zp
            if isinstance(m, torch.nn.Linear):
                in_features = m.in_features
                out_features = m.out_features
            elif is_transformers_imported() and isinstance(m, transformers.Conv1D):
                in_features = m.weight.shape[0]
                out_features = m.weight.shape[1]
                int_weight = int_weight.t_().contiguous()
                scale = scale.t_().contiguous()
                zp = zp.t_().contiguous() if zp is not None else zp

            new_module = INCWeightOnlyLinear(
                in_features,
                out_features,
                dtype=dtype,
                bits=bits,
                group_size=group_size,
                zp=zp is not None,
                bias=m.bias is not None,
                use_optimum_format=use_optimum_format,
                device=device,
            )
            new_module.pack(int_weight, scale, zp, m.bias)

            if use_layer_wise:
                m = m.to_empty(device=torch.device("meta"))
            if name == "":
                return new_module
            else:
                set_module(model, name, new_module)
            # Move modules back to the model device layer-by-layer
            if not use_layer_wise:
                m.to(model_device)
                new_module.to(model_device)
        if not use_layer_wise:
            model.to(model_device)
        return model



# Copyright (c) 2023-2024 Microsoft Corporation and Intel Corporation

# This code is based on Microsoft Corporation's DeepSpeed library and
# the accelerators implementation in this library. It has been modified
# from its original forms to simplify and adapt it for use in
# the Intel® Neural Compressor.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE: The design adapted from:
# https://github.com/microsoft/DeepSpeed/blob/master/accelerator/abstract_accelerator.py.
"""Auto Accelerator Module."""


# To keep it simply, only add the APIs we need.

import gc
import os
from abc import ABC, abstractmethod
from enum import Enum, auto
from functools import lru_cache
from typing import Any, Callable, List

import torch

from neural_compressor.common.utils import LazyImport, logger

htcore = LazyImport("habana_frameworks.torch.core")

PRIORITY_HPU = 100
PRIORITY_XPU = 95
PRIORITY_CUDA = 90
PRIORITY_CPU = 80


class INCAcceleratorType(Enum):
    CUDA = auto()
    CPU = auto()
    EXTERNAL_ACCELERATOR_MAX = auto()  # differentiate between external to intel accelerators
    XPU = auto()
    GAUDI_MIN = auto()  # differentiate between any gaudi to xpu
    GAUDI2 = auto()
    GAUDI3 = auto()


class AcceleratorRegistry:
    """Accelerator Registry."""

    registered_accelerators = {}

    @classmethod
    def register_accelerator_impl(cls, name: str, priority: float = 0):
        """Register new accelerator implementation.

        Usage example:
            @AcceleratorRegistry.register_accelerator(name="cpu", priority=100)
            class CPU_Accelerator:
                ...

        Args:
            name: the accelerator name.
            priority: priority: the priority of the accelerator. A larger number indicates a higher priority,
        """

        def decorator(accelerator_cls):
            if accelerator_cls.is_available():
                cls.registered_accelerators.setdefault(name, {})
                cls.registered_accelerators[name] = (accelerator_cls, priority)
            return accelerator_cls

        return decorator

    @classmethod
    def get_sorted_accelerators(cls) -> List["Auto_Accelerator"]:
        """Get registered accelerators sorted by priority."""
        accelerator_pairs = cls.registered_accelerators.values()
        sorted_accelerators_pairs = sorted(accelerator_pairs, key=lambda x: x[1], reverse=True)
        sorted_accelerators = [pair[0] for pair in sorted_accelerators_pairs]
        return sorted_accelerators

    @classmethod
    def get_accelerator_cls_by_name(cls, name: str) -> "Auto_Accelerator":
        """Get accelerator by name."""
        accelerator_cls, _ = cls.registered_accelerators.get(name, (None, None))
        return accelerator_cls


accelerator_registry = AcceleratorRegistry()


def register_accelerator(name: str, priority: float = 0) -> Callable[..., Any]:
    """Register new accelerator.

    Usage example:
        @register_accelerator(name="cuda", priority=100)
        class CUDA_Accelerator:
            ...

    Args:
        name: the accelerator name.
        priority: the priority of the accelerator. A larger number indicates a higher priority,
    """
    return accelerator_registry.register_accelerator_impl(name=name, priority=priority)


class Auto_Accelerator(ABC):  # pragma: no cover
    """Auto Accelerator Base class."""

    @classmethod
    @abstractmethod
    def is_available(cls) -> bool:
        """Check if the accelerator is available."""
        pass

    @abstractmethod
    def name(self) -> str:
        """Get the accelerator name."""
        pass

    @abstractmethod
    def device_name(self, device_indx) -> str:
        """Get the device name."""
        pass

    @abstractmethod
    def set_device(self, device_index):
        """Set the device."""
        pass

    @abstractmethod
    def current_device(self):
        """Get the current device."""
        pass

    @abstractmethod
    def current_device_name(self):
        """Get the current device name."""
        pass

    @abstractmethod
    def device(self, device_index=None):
        """Get the device."""
        pass

    @abstractmethod
    def empty_cache(self):
        """Empty the cache."""
        pass

    @abstractmethod
    def synchronize(self):
        """Synchronize the accelerator."""
        pass

    @abstractmethod
    def get_inc_accelerator_type(self) -> INCAcceleratorType:
        """Synchronize the accelerator."""
        pass


@register_accelerator(name="cpu", priority=PRIORITY_CPU)
class CPU_Accelerator(Auto_Accelerator):
    """CPU Accelerator."""

    def __init__(self) -> None:
        """Initialize CPU Accelerator."""
        self._name = "cpu"

    def name(self) -> str:
        """Get the accelerator name."""
        return self._name

    @classmethod
    def is_available(cls) -> bool:
        """Always return True."""
        return True

    def device_name(self, device_indx) -> str:
        """Get the device name."""
        return "cpu"

    def set_device(self, device_index):
        """Do nothing."""
        pass

    def current_device(self):
        """Get the current device."""
        return "cpu"

    def current_device_name(self):
        """Get the current device name."""
        return "cpu"

    def device(self, device_index=None):
        """Do nothing."""
        pass

    def empty_cache(self):
        """Do nothing."""
        gc.collect()

    def synchronize(self):
        """Do nothing."""
        pass

    def get_inc_accelerator_type(self):
        return INCAcceleratorType.CPU


@register_accelerator(name="cuda", priority=PRIORITY_CUDA)
class CUDA_Accelerator(Auto_Accelerator):  # pragma: no cover
    """CUDA Accelerator."""

    def __init__(self) -> None:
        """Initialize CUDA Accelerator."""
        self._name = "cuda"

    def name(self) -> str:
        """Get the accelerator name."""
        return self._name

    @classmethod
    def is_available(cls) -> bool:
        """Check if the 'cuda' device is available."""
        return torch.cuda.is_available()

    def device_name(self, device_indx) -> str:
        """Returns the name of the 'cuda' device with the given index."""
        if device_indx is None:
            return "cuda"
        return f"cuda:{device_indx}"

    def synchronize(self):
        """Synchronizes the 'cuda' device."""
        return torch.cuda.synchronize()

    def set_device(self, device_index):
        """Sets the current 'cuda' device to the one with the given index."""
        return torch.cuda.set_device(device_index)

    def current_device(self):
        """Returns the index of the current 'cuda' device."""
        return torch.cuda.current_device()

    def current_device_name(self):
        """Returns the name of the current 'cuda' device."""
        return "cuda:{}".format(torch.cuda.current_device())

    def device(self, device_index=None):
        """Returns a torch.device object for the 'cuda' device with the given index."""
        return torch.cuda.device(device_index)

    def empty_cache(self):
        """Empties the cuda cache."""
        return torch.cuda.empty_cache()

    def get_inc_accelerator_type(self):
        return INCAcceleratorType.CUDA


@register_accelerator(name="xpu", priority=PRIORITY_XPU)
class XPU_Accelerator(Auto_Accelerator):  # pragma: no cover
    """XPU Accelerator."""

    def __init__(self) -> None:
        """Initialize XPU Accelerator."""
        self._name = "xpu"

    def name(self) -> str:
        """Get the accelerator name."""
        return self._name

    @classmethod
    def is_available(cls) -> bool:
        """Checks if the 'xpu' device is available.

        Returns:
            bool: True if the 'xpu' device is available, False otherwise.
        """
        if hasattr(torch, "xpu") and torch.xpu.is_available():
            return True
        else:
            return False

    def device_name(self, device_indx) -> str:
        """Returns the name of the 'xpu' device with the given index.

        Args:
            device_indx (int): The index of the 'xpu' device.

        Returns:
            str: The name of the 'xpu' device.
        """
        if device_indx is None:
            return "xpu"
        return f"xpu:{device_indx}"

    def synchronize(self):
        """Synchronizes the 'xpu' device."""
        return torch.xpu.synchronize()

    def set_device(self, device_index):
        """Sets the current 'xpu' device to the one with the given index.

        Args:
            device_index (int): The index of the 'xpu' device.
        """
        return torch.xpu.set_device(device_index)

    def current_device(self):
        """Returns the index of the current 'xpu' device.

        Returns:
            int: The index of the current 'xpu' device.
        """
        return torch.xpu.current_device()

    def current_device_name(self):
        """Returns the name of the current 'xpu' device.

        Returns:
            str: The name of the current 'xpu' device.
        """
        return "xpu:{}".format(torch.xpu.current_device())

    def device(self, device_index=None):
        """Returns a torch.device object for the 'xpu' device with the given index.

        Args:
            device_index (int, optional): The index of the 'xpu' device. Defaults to None.

        Returns:
            torch.device: The torch.device object for the 'xpu' device.
        """
        return torch.xpu.device(device_index)

    def empty_cache(self):
        """Empties the xpu cache."""
        return torch.xpu.empty_cache()

    def get_inc_accelerator_type(self):
        return INCAcceleratorType.XPU


@register_accelerator(name="hpu", priority=PRIORITY_HPU)
class HPU_Accelerator(Auto_Accelerator):  # pragma: no cover
    """HPU Accelerator."""

    def __init__(self) -> None:
        """Initialize HPU Accelerator."""
        self._name = "hpu"

    def name(self) -> str:
        """Get the accelerator name."""
        return self._name

    @classmethod
    def is_available(cls) -> bool:
        """Checks if the 'hpu' device is available."""
        from .environ import is_hpex_available

        if is_hpex_available():
            return torch.hpu.is_available()
        else:
            return False

    def device_name(self, device_indx) -> str:
        """Returns the name of the 'hpu' device with the given index."""
        if device_indx is None:
            return "hpu"
        return f"hpu:{device_indx}"

    def synchronize(self):
        """Synchronizes the 'hpu' device."""
        logger.debug("Calling `htcore.mark_step()` and `torch.hpu.synchronize()`.")
        htcore.mark_step()
        return torch.hpu.synchronize()

    def set_device(self, device_index):
        """Sets the current 'hpu' device to the one with the given index."""
        try:
            torch.hpu.set_device(device_index)
        except Exception as e:
            logger.warning(e)

    def current_device(self):
        """Returns the index of the current 'hpu' device."""
        return torch.hpu.current_device()

    def current_device_name(self):
        """Returns the name of the current 'hpu' device."""
        return "hpu:{}".format(torch.hpu.current_device())

    def device(self, device_index=None):
        """Returns a torch.device object for the 'hpu' device with the given index."""
        return torch.hpu.device(device_index)

    def empty_cache(self):
        """Empties the hpu cache."""
        try:
            torch.hpu.empty_cache()
        except Exception as e:
            logger.warning(e)

    def get_inc_accelerator_type(self):
        import habana_frameworks.torch.utils.experimental as htexp

        hpu_device_type = htexp._get_device_type()
        if hpu_device_type == htexp.synDeviceType.synDeviceGaudi2:
            return INCAcceleratorType.GAUDI2
        elif hpu_device_type == htexp.synDeviceType.synDeviceGaudi3:
            return INCAcceleratorType.GAUDI3
        else:
            raise ValueError("Unexpected Device Type for HPU accelerator - {}".format(hpu_device_type))


@lru_cache()
def auto_detect_accelerator(device_name="auto") -> Auto_Accelerator:
    """Automatically detects and selects the appropriate accelerator.

    Force use the cpu on node has both cpu and gpu: `INC_TARGET_DEVICE=cpu` python main.py ...
    The `INC_TARGET_DEVICE` is case insensitive.
    The environment variable `INC_TARGET_DEVICE` has higher priority than the `device_name`.
    TODO: refine the docs and logic later
    """
    # 1. Get the device setting from environment variable `INC_TARGET_DEVICE`.
    INC_TARGET_DEVICE = os.environ.get("INC_TARGET_DEVICE", None)
    if INC_TARGET_DEVICE:
        INC_TARGET_DEVICE = INC_TARGET_DEVICE.lower()
    # 2. If the `INC_TARGET_DEVICE` is set and the accelerator is available, use it.
    if INC_TARGET_DEVICE and accelerator_registry.get_accelerator_cls_by_name(INC_TARGET_DEVICE) is not None:
        logger.warning("Force use %s accelerator.", INC_TARGET_DEVICE)
        return accelerator_registry.get_accelerator_cls_by_name(INC_TARGET_DEVICE)()
    # 3. If the `device_name` is set and the accelerator is available, use it.
    if device_name != "auto":
        if accelerator_registry.get_accelerator_cls_by_name(device_name) is not None:
            accelerator_cls = accelerator_registry.get_accelerator_cls_by_name(device_name)
            logger.warning("Selected accelerator %s by device_name.", accelerator_cls.__name__)
            return accelerator_cls()
        else:
            logger.warning("The device name %s is not supported, use auto detect instead.", device_name)
    # 4. Select the accelerator by priority.
    for accelerator_cls in accelerator_registry.get_sorted_accelerators():
        if accelerator_cls.is_available():
            logger.warning("Auto detect accelerator: %s.", accelerator_cls.__name__)
            accelerator = accelerator_cls()
            return accelerator


# Force use cpu accelerator even if cuda is available.
# INC_TARGET_DEVICE = "cpu" python ...
# or
# INC_TARGET_DEVICE = "CPU" python ...
# or
# CUDA_VISIBLE_DEVICES="" python ...


def is_any_gaudi_accelerator(acc_type: INCAcceleratorType):
    return acc_type.value > INCAcceleratorType.GAUDI_MIN.value



"""External distributed collective function import helpers."""

# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# File contains logic for importing from external source distributed functions, such as we import them only when needed and possible,
# and protect from import errors in case the external source is not installed.

try:
    from vllm.distributed import (
        tensor_model_parallel_all_gather,
        tensor_model_parallel_all_reduce,
    )
except ImportError:
    try:
        from sglang.srt.distributed import (
            tensor_model_parallel_all_gather,
            tensor_model_parallel_all_reduce,
        )
    except ImportError:
        tensor_model_parallel_all_gather = None
        tensor_model_parallel_all_reduce = None

def get_external_column_parallel_collective_func():
    """Return the column-parallel all-gather collective from external runtime.

    Returns:
        Callable: The tensor-model-parallel all-gather function.
    """
    assert (
        tensor_model_parallel_all_gather is not None
    ), "Could not import function tensor_model_parallel_all_gather from external source"
    return tensor_model_parallel_all_gather

def get_external_row_parallel_collective_func():
    """Return the row-parallel all-reduce collective from external runtime.

    Returns:
        Callable: The tensor-model-parallel all-reduce function.
    """
    assert (
        tensor_model_parallel_all_reduce is not None
    ), "Could not import function tensor_model_parallel_all_reduce from external source"
    return tensor_model_parallel_all_reduce



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import torch
import tqdm

from .quarot_utils import get_hadK

# This code implements rotations to the model, and is based on the paper below:
# "QuaRot: Outlier-Free 4-Bit Inference in Rotated LLMs". See https://github.com/spcl/QuaRot/tree/main.
# The script rotates weights/activations with Hadamard matrices to reduce outliers and improve quantization.
# Tested on llama2-7b, llama2-13b, llama3-8b.
# The code is compatible with GPTQ (rotation is applied first, followed by GPTQ on the rotated model).
# Rotating the weights and the values is done offline, and does not decrease inference speed.
# Rotating the MLP layers is partially online, which adds computational overhead.
# To rotate, call the function "rotate(model, args)".
# Calling the function rotates the weights. To also rotate the values and/or MLP, pass the arguments:
# args.rotate_values and/or args.rotate_mlp.

DEV = "hpu"


def fuse_ln_linear(layernorm: torch.nn.Module, linear_layers: typing.Iterable[torch.nn.Linear]) -> None:
    """Fuse the linear operations in Layernorm into the adjacent linear blocks."""
    for linear in linear_layers:
        linear_dtype = linear.weight.dtype
        W_ = linear.weight.data.double()
        linear.weight.data = (W_ * layernorm.weight.double()).to(linear_dtype)
        if hasattr(layernorm, "bias"):
            if linear.bias is None:
                linear.bias = torch.nn.Parameter(torch.zeros(linear.out_features, dtype=torch.float64))
            linear.bias.data = linear.bias.data.double() + torch.matmul(W_, layernorm.bias.double())
            linear.bias.data = linear.bias.data.to(linear_dtype)


def fuse_layer_norms(model) -> None:
    # Embedding fusion
    W = model.model.embed_tokens
    W_ = W.weight.data.double()
    W.weight.data = (W_ - W_.mean(dim=-1, keepdim=True)).to(W.weight.data.dtype)
    # Fuse the linear operations in Layernorm into the adjacent linear blocks.
    layers = [layer for layer in model.model.layers]
    for layer in layers:
        fuse_ln_linear(layer.post_attention_layernorm, [layer.mlp.up_proj, layer.mlp.gate_proj])
        fuse_ln_linear(layer.input_layernorm, [layer.self_attn.q_proj, layer.self_attn.k_proj, layer.self_attn.v_proj])
        W_norm = layer.post_attention_layernorm.weight.data
        # We moved the parameters to the weights matrices, thus we replace them with ones:
        layer.post_attention_layernorm.weight.data = torch.ones_like(W_norm)
        W_norm = layer.input_layernorm.weight.data
        layer.input_layernorm.weight.data = torch.ones_like(W_norm)
    fuse_ln_linear(model.model.norm, [model.lm_head])
    W_norm = model.model.norm.weight.data
    model.model.norm.weight.data = torch.ones_like(W_norm)


def matmul_hadU(X, transpose=False):
    n = X.shape[-1]
    hadK, K = get_hadK(n, transpose)
    input = X.clone().view(-1, n, 1)
    output = input.clone()
    while input.shape[1] > K:
        input = input.view(input.shape[0], input.shape[1] // 2, 2, input.shape[2])
        output = output.view(input.shape)
        output[:, :, 0, :] = input[:, :, 0, :] + input[:, :, 1, :]
        output[:, :, 1, :] = input[:, :, 0, :] - input[:, :, 1, :]
        output = output.view(input.shape[0], input.shape[1], -1)
        (input, output) = (output, input)
    del output
    if K > 1:
        input = hadK.view(1, K, K).to(input) @ input
    return input.view(X.shape) / torch.tensor(n).sqrt()


def get_orthogonal_matrix(size, random=False):
    # See https://cornell-relaxml.github.io/quip-sharp/
    if random:
        Q = torch.randint(low=0, high=2, size=(size,)).to(torch.float64)
        Q = Q * 2 - 1
    else:
        Q = torch.ones(size, dtype=torch.float64)
    Q = torch.diag(Q)
    return matmul_hadU(Q)


def get_kron_hadamard(size):
    hadK, K = get_hadK(size)
    normalization = torch.sqrt(torch.tensor(K))
    hadK = hadK / normalization
    p_hadamard = get_orthogonal_matrix(int(size / K), random=False)
    return torch.kron(p_hadamard, hadK)


def rotate_head(model, Q: torch.Tensor) -> None:
    W = model.lm_head
    W_ = W.weight.data.to(device=DEV, dtype=torch.float64)
    W.weight.data = torch.matmul(W_, Q).to(device="cpu", dtype=model.dtype)


def rotate_embeddings(model, Q: torch.Tensor) -> None:
    for W in [model.model.embed_tokens]:
        W_ = W.weight.data.to(device=DEV, dtype=torch.float64)
        W.weight.data = torch.matmul(W_, Q).to(device="cpu", dtype=model.dtype)


def rotate_attention_inputs(layer, Q) -> None:
    # Rotate the WQ, WK and WV matrices of the self-attention layer.
    for W in [layer.self_attn.q_proj, layer.self_attn.k_proj, layer.self_attn.v_proj]:
        dtype = W.weight.dtype
        W_ = W.weight.to(device="hpu", dtype=torch.float64)
        W.weight.data = torch.matmul(W_, Q).to(device="cpu", dtype=dtype)


def rotate_attention_output(layer, Q) -> None:
    W = layer.self_attn.o_proj
    dtype = W.weight.data.dtype
    W_ = W.weight.data.to(device=DEV, dtype=torch.float64)
    W.weight.data = torch.matmul(Q.T, W_).to(device="cpu", dtype=dtype)
    if W.bias is not None:
        b = W.bias.data.to(device=DEV, dtype=torch.float64)
        W.bias.data = torch.matmul(Q.T, b).to(device="cpu", dtype=dtype)


def rotate_mlp_input(layer, Q) -> None:
    mlp_inputs = [layer.mlp.up_proj, layer.mlp.gate_proj]
    for W in mlp_inputs:
        dtype = W.weight.dtype
        W_ = W.weight.data.to(device=DEV, dtype=torch.float64)
        W.weight.data = torch.matmul(W_, Q).to(device="cpu", dtype=dtype)


def rotation_mlp_pre_hook(module, input):
    # This add online rotation in the mlp layer.
    rotated_input = torch.matmul(input[0], module.H_down)
    return rotated_input


def add_forward_mlp_wrapper(model) -> None:
    # This add online rotation in the mlp layer if and is used on you load a rotated model.
    config = model.config
    H_down = get_kron_hadamard(config.intermediate_size).to(device=DEV, dtype=torch.float64)
    layers = [layer for layer in model.model.layers]
    for layer in layers:
        W = layer.mlp.down_proj
        W.register_buffer("H_down", H_down.to(device="cpu", dtype=model.dtype))
        hook_handle = W.register_forward_pre_hook(rotation_mlp_pre_hook)


def rotate_mlp_output(layer, config, Q) -> None:
    # This function rotates the weight matrices at the output of the mlp layer.
    # when rotating the activations within the layer, the function adds a hook which perform online rotation.
    W = layer.mlp.down_proj
    dtype = W.weight.data.dtype
    W_ = W.weight.data.to(device=DEV, dtype=torch.float64)
    W.weight.data = torch.matmul(Q.T, W_).to(device="cpu", dtype=dtype)
    if hasattr(config, "rotate_mlp") and config.rotate_mlp:
        H_down = get_kron_hadamard(config.intermediate_size).to(device=DEV, dtype=torch.float64)
        W_ = W.weight.data.to(device=DEV, dtype=torch.float64)
        W.weight.data = torch.matmul(W_, H_down).to(device="cpu", dtype=dtype)
        W.register_buffer("H_down", H_down.to(device="cpu", dtype=dtype))
        hook_handle = W.register_forward_pre_hook(rotation_mlp_pre_hook)
    if W.bias is not None:
        b = W.bias.data.to(device=DEV, dtype=torch.float64)
        W.bias.data = torch.matmul(Q.T, b).to(device="cpu", dtype=dtype)


def rotate_ov_proj(layer) -> None:
    v_proj = layer.self_attn.v_proj
    o_proj = layer.self_attn.o_proj
    dtype = o_proj.weight.dtype
    num_q_heads = layer.self_attn.num_heads
    num_kv_heads = layer.self_attn.num_key_value_heads
    head_dim = layer.self_attn.head_dim
    H_head = get_orthogonal_matrix(head_dim, random=False).to(device=DEV, dtype=torch.float64)
    I_v = torch.eye(num_kv_heads).to(device=DEV, dtype=torch.float64)
    I_out = torch.eye(num_q_heads).to(device=DEV, dtype=torch.float64)
    H_v = torch.kron(I_v, H_head)
    H_out = torch.kron(I_out, H_head)
    W_ = v_proj.weight.data.to(device=DEV, dtype=torch.float64)
    v_proj.weight.data = torch.matmul(H_v.T, W_).to(device="cpu", dtype=dtype)
    W_ = o_proj.weight.data.to(device=DEV, dtype=torch.float64)
    o_proj.weight.data = torch.matmul(W_, H_out).to(device="cpu", dtype=dtype)


def rotate_model(model) -> None:
    Q = get_orthogonal_matrix(model.config.hidden_size, random=True).to(device=DEV, dtype=torch.float64)
    rotate_embeddings(model, Q)
    rotate_head(model, Q)
    layers = [layer for layer in model.model.layers]
    for idx, layer in enumerate(tqdm.tqdm(layers, unit="layer", desc="Rotating")):
        rotate_attention_inputs(layers[idx], Q)
        rotate_attention_output(layers[idx], Q)
        rotate_mlp_input(layers[idx], Q)
        rotate_mlp_output(layers[idx], model.config, Q)
        if hasattr(model.config, "rotate_values") and model.config.rotate_values:
            rotate_ov_proj(layers[idx])


def rotate(model, args) -> None:
    model.config.rotate_weights = True
    if args.rotate_mlp:
        model.config.rotate_mlp = True
    if args.rotate_values:
        model.config.rotate_values = True
    fuse_layer_norms(model)
    rotate_model(model)



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor Transformers-like Config."""

import os
from typing import Any, Dict, Tuple, Union

from neural_compressor.common.utils import LazyImport, logger

torch = LazyImport("torch")
transformers = LazyImport("transformers")

QUANT_CONFIG = "quantize_config.json"

if transformers.__version__ >= "4.32.0":
    from transformers.utils.quantization_config import QuantizationConfigMixin

    QuantizationConfig = QuantizationConfigMixin
else:
    from transformers import PretrainedConfig

    QuantizationConfig = PretrainedConfig
from enum import Enum


class QuantizationMethod(str, Enum):
    GPTQ = "gptq"
    RTN = "rtn"
    AWQ = "awq"
    TEQ = "teq"


class INCQuantizationConfigMixin(QuantizationConfig):
    """Mixin class for quantization config."""

    def update(self, **kwargs):
        """Updates attributes of this class instance with attributes from `kwargs` if they match existing atributtes,
        returning all the unused kwargs.

        Args:
            kwargs (`Dict[str, Any]`):
                Dictionary of attributes to tentatively update this class.

        Returns:
            `Dict[str, Any]`: Dictionary containing all the key-value pairs that were not used to update the instance.
        """
        to_remove = []
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
                to_remove.append(key)

        # Remove all the attributes that were updated, without modifying the input dict
        unused_kwargs = {key: value for key, value in kwargs.items() if key not in to_remove}
        return unused_kwargs

    def post_init_cpu(self):
        r"""Safety checker that arguments are correct."""

        if self.compute_dtype is not None and self.compute_dtype not in [
            "fp32",
            "bf16",
            "int8",
        ]:
            raise ValueError("compute_dtype must be 'fp32', 'bf16', 'int8'.")
        elif self.compute_dtype is None:
            self.compute_dtype = "fp32"

        if self.bits is None:
            self.bits = 4
        elif self.bits is not None and self.bits not in [4, 8]:
            raise ValueError(f"Only support quantization to [4, 8] bits but found {self.bits}")

        if self.scale_dtype is not None and self.scale_dtype not in ["fp32", "bf16", "fp16"]:
            raise ValueError("scale_dtype must be a string in 'fp32', 'bf16' ")
        elif self.scale_dtype is None:
            self.scale_dtype = "fp32"

        if not isinstance(self.group_size, int):
            raise ValueError("group_size must be a int")

        if not isinstance(self.scheme, str):
            raise ValueError("scheme must be a string")

    def post_init_xpu(self):
        r"""
        Safety checker that arguments are correct - also replaces some NoneType arguments with their default values.
        """

        if self.compute_dtype is not None and self.compute_dtype not in ["fp16"]:
            raise ValueError("compute_dtype must be 'fp16'.")
        elif self.compute_dtype is None:
            self.compute_dtype = "fp16"

        if self.bits is None:
            self.bits = 4
        elif self.bits not in [4]:
            raise ValueError(f"Only support quantization to [4] bits but found {self.bits}")

        if self.weight_dtype is None:
            self.weight_dtype = "int4_fullrange"
        elif self.weight_dtype == "int4":
            self.weight_dtype = "int4_fullrange"
        elif self.weight_dtype not in [
            "int4_fullrange",
        ]:
            raise ValueError(f"weight_dtype must be a string in 'int4_fullrange', but get {self.weight_dtype}.")

        if self.scale_dtype is not None and self.scale_dtype not in ["fp16"]:
            raise ValueError("scale_dtype must be a string in 'fp16'")
        elif self.scale_dtype is None:
            self.scale_dtype = "fp16"

        if not isinstance(self.group_size, int):
            raise ValueError("group_size must be a int")

        if self.scheme not in ["sym"]:
            raise ValueError("scheme: {} is not support, only support 'sym' now!".format(self.scheme))

    def to_json_file(self, json_file_path: Union[str, os.PathLike], use_diff: bool = True):
        """Save this instance to a JSON file.

        Args:
            json_file_path (`str` or `os.PathLike`):
                Path to the JSON file in which this configuration instance's parameters will be saved.
        """
        # set tokenizer to None due to it doesn't support write to json
        if hasattr(self, "tokenizer"):
            self.tokenizer = None
        if hasattr(self, "calib_dataloader"):
            self.calib_dataloader = None
        with open(json_file_path, "w", encoding="utf-8") as writer:
            writer.write(self.to_json_string(use_diff=use_diff))

    def remove_redundant_parameters(self):
        remove_parameters = [
            "calib_dataloader",
            "dataset",
            "calib_func",
            "calib_iters",
            "calib_len",
            "mse_range",
            "scheme",
            "tokenizer",
            "use_layer_wise",
            "blocksize",
            "nsamples",
            "max_input_length",
            "static_groups",
            "lr",
            "minmax_lr",
            "iters",
            "use_quant_input",
            "device",
            "calib_dataset",
            "calib_pad_val",
            "calib_shuffle",
            "calib_padding",
            "example_inputs",
            "excluded_precisions",
            "op_name_dict",
            "op_type_dict",
            "train_dataloader",
            "train_func",
            "train_iters",
            "train_len",
            "train_padding",
            "train_dataset",
            "train_pad_val",
            "train_shuffle",
            "train_batch_size",
        ]
        for parameter in remove_parameters:
            if hasattr(self, parameter):
                delattr(self, parameter)

    def save_pretrained(
        self,
        save_directory: Union[str, os.PathLike],
        push_to_hub: bool = False,
        **kwargs,
    ):
        """Save a configuration object to the directory `save_directory`, so that it can be re-loaded using the
        [`~PretrainedConfig.from_pretrained`] class method.

        Args:
            save_directory (`str` or `os.PathLike`):
                Directory where the configuration JSON file will be saved (will be created if it does not exist).
            push_to_hub (`bool`, *optional*, defaults to `False`):
                Whether or not to push your model to the Hugging Face model hub after saving it. You can specify the
                repository you want to push to with `repo_id` (will default to the name of `save_directory` in your
                namespace).
            kwargs (`Dict[str, Any]`, *optional*):
                Additional key word arguments passed along to the [`~utils.PushToHubMixin.push_to_hub`] method.
        """
        if os.path.isfile(save_directory):
            raise AssertionError(f"Provided path ({save_directory}) should be a directory, not a file")

        os.makedirs(save_directory, exist_ok=True)

        if push_to_hub:
            commit_message = kwargs.pop("commit_message", None)
            repo_id = kwargs.pop("repo_id", save_directory.split(os.path.sep)[-1])
            repo_id = self._create_repo(repo_id, **kwargs)
            files_timestamps = self._get_files_timestamps(save_directory)

        # If we save using the predefined names, we can load using `from_pretrained`
        output_config_file = os.path.join(save_directory, QUANT_CONFIG)

        self.to_json_file(output_config_file, use_diff=False)
        logger.info(f"Configuration saved in {output_config_file}")

        if push_to_hub:
            self._upload_modified_files(
                save_directory,
                repo_id,
                files_timestamps,
                commit_message=commit_message,
                token=kwargs.get("token", None),
            )

    @classmethod
    def get_config_dict(
        cls, pretrained_model_name_or_path: Union[str, os.PathLike], **kwargs
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        cf = kwargs.pop("_configuration_file", QUANT_CONFIG)
        return super().get_config_dict(pretrained_model_name_or_path, _configuration_file=cf, **kwargs)


class RtnConfig(INCQuantizationConfigMixin):
    def __init__(
        self,
        bits: int = 4,
        group_size: int = 32,
        compute_dtype: Any = None,
        scale_dtype: Any = None,
        sym: bool = True,
        use_layer_wise: bool = None,
        quant_lm_head: bool = False,
        **kwargs,
    ):
        self.quant_method = QuantizationMethod.RTN
        self.bits = bits
        self.compute_dtype = compute_dtype
        self.weight_dtype = "int4" if self.bits == 4 else "int8"
        self.scale_dtype = scale_dtype
        self.group_size = group_size
        self.use_layer_wise = use_layer_wise
        self.model_path = kwargs.get("model_path", "")
        self.sym = sym
        self.scheme = "sym" if self.sym else "asym"

        # "transformer.output_layer" for chatglm series model.
        # "embed_out" for dolly v2 series model.
        self.quant_lm_head = quant_lm_head
        self.modules_to_not_convert = kwargs.get(
            "modules_to_not_convert", ["lm_head", "transformer.output_layer", "embed_out"]
        )
        if self.quant_lm_head:
            self.modules_to_not_convert = []
        self.device = kwargs.get("device", "auto")

    def to_diff_dict(self) -> Dict[str, Any]:
        """Removes all attributes from config which correspond to the default config attributes
        for better readability and serializes to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        config_dict = self.to_dict()

        # get the default config dict
        default_config_dict = RtnConfig().to_dict()

        serializable_config_dict = {}

        # only serialize values that differ from the default config
        for key, value in config_dict.items():
            if value != default_config_dict[key]:
                serializable_config_dict[key] = value

        return serializable_config_dict


class GPTQConfig(INCQuantizationConfigMixin):
    def __init__(
        self,
        bits: int = 4,
        tokenizer: Any = None,
        dataset: str = "NeelNanda/pile-10k",
        batch_size: int = 8,
        group_size: int = 32,
        compute_dtype: Any = None,
        scale_dtype: Any = None,
        sym: bool = True,
        blocksize: int = 128,
        damp_percent: float = 0.1,
        desc_act: bool = False,
        n_samples: int = 128,
        seq_len: int = 2048,
        static_groups: bool = False,
        use_mse_search: bool = False,
        true_sequential: bool = False,
        use_layer_wise: bool = None,
        quant_lm_head: bool = False,
        **kwargs,
    ):

        self.quant_method = QuantizationMethod.GPTQ
        self.bits = bits
        self.tokenizer = tokenizer
        self.dataset = dataset
        self.batch_size = batch_size
        self.compute_dtype = compute_dtype
        self.weight_dtype = "int4" if self.bits == 4 else "int8"
        self.scale_dtype = scale_dtype
        self.sym = sym
        self.blocksize = blocksize
        self.n_samples = n_samples
        self.group_size = group_size
        self.damp_percent = damp_percent
        self.desc_act = desc_act
        self.static_groups = static_groups
        self.use_mse_search = use_mse_search
        self.true_sequential = true_sequential
        self.use_layer_wise = use_layer_wise
        self.model_path = kwargs.get("model_path", "")
        self.seq_len = seq_len
        self.quant_lm_head = quant_lm_head
        self.modules_to_not_convert = kwargs.get(
            "modules_to_not_convert", ["lm_head", "transformer.output_layer", "embed_out"]
        )
        if self.quant_lm_head:
            self.modules_to_not_convert = []
        self.device = kwargs.get("device", "auto")
        self.scheme = "sym" if self.sym else "asym"

        if isinstance(compute_dtype, torch.dtype):
            self.compute_dtype = compute_dtype
        else:
            self.compute_dtype = compute_dtype

        if isinstance(scale_dtype, torch.dtype):
            self.scale_dtype = scale_dtype
        else:
            self.scale_dtype = scale_dtype

        self.post_init_gptq()

    def post_init_gptq(self):
        r"""Safety checker that arguments are correct."""

        if self.bits not in [4, 8]:
            raise ValueError(f"Only support quantization to [4, 8] bits but found {self.bits}")

        if not (0 < self.damp_percent < 1):
            raise ValueError("damp_percent must between 0 and 1.")

    def to_diff_dict(self) -> Dict[str, Any]:
        """Removes all attributes from config which correspond to the default config attributes
        for better readability and serializes to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        config_dict = self.to_dict()

        # get the default config dict
        default_config_dict = GPTQConfig().to_dict()

        serializable_config_dict = {}

        # only serialize values that differ from the default config
        for key, value in config_dict.items():
            if value != default_config_dict[key]:
                serializable_config_dict[key] = value

        return serializable_config_dict


class AwqConfig(INCQuantizationConfigMixin):
    def __init__(
        self,
        bits: int = 4,
        tokenizer: Any = None,
        dataset: str = "NeelNanda/pile-10k",
        group_size: int = 32,
        compute_dtype: Any = None,
        weight_dtype: Any = None,
        scale_dtype: Any = None,
        use_layer_wise: bool = None,
        n_samples: int = 128,
        seq_len: int = 2048,
        auto_scale: bool = True,
        auto_clip: bool = True,
        zero_point: bool = True,
        absorb_layer_dict: dict = {},
        quant_lm_head: bool = False,
        backend: str = None,
        **kwargs,
    ):
        self.quant_method = QuantizationMethod.AWQ
        self.bits = bits
        self.tokenizer = tokenizer
        self.dataset = dataset
        self.compute_dtype = compute_dtype
        self.weight_dtype = "int4" if self.bits == 4 else "int8"
        self.scale_dtype = scale_dtype
        self.group_size = group_size
        self.zero_point = zero_point
        self.auto_scale = auto_scale
        self.auto_clip = auto_clip
        self.use_layer_wise = use_layer_wise
        self.n_samples = n_samples
        self.seq_len = seq_len
        self.absorb_layer_dict = absorb_layer_dict
        self.quant_lm_head = quant_lm_head
        self.backend = backend
        self.modules_to_not_convert = kwargs.get(
            "modules_to_not_convert", ["lm_head", "transformer.output_layer", "embed_out"]
        )
        if self.quant_lm_head:
            self.modules_to_not_convert = []
        self.device = kwargs.get("device", "auto")
        self.scheme = "asym" if self.zero_point else "sym"
        self.sym = True if not self.zero_point else False
        self.batch_size = kwargs.pop("batch_size", 8)

    def to_diff_dict(self) -> Dict[str, Any]:
        """Removes all attributes from config which correspond to the default config attributes
        for better readability and serializes to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        config_dict = self.to_dict()

        # get the default config dict
        default_config_dict = AwqConfig().to_dict()

        serializable_config_dict = {}

        # only serialize values that differ from the default config
        for key, value in config_dict.items():
            if value != default_config_dict[key]:
                serializable_config_dict[key] = value

        return serializable_config_dict


class TeqConfig(INCQuantizationConfigMixin):
    def __init__(
        self,
        bits: int = 4,
        tokenizer: Any = None,
        dataset: str = "NeelNanda/pile-10k",
        group_size: int = 32,
        compute_dtype: Any = None,
        weight_dtype: Any = None,
        scale_dtype: Any = None,
        use_layer_wise: bool = None,
        n_samples: int = 128,
        seq_len: int = 2048,
        sym: bool = True,
        folding: bool = False,  # TODO, add folding support for transformers >= 4.55.2
        absorb_layer_dict: dict = {},
        quant_lm_head: bool = False,
        **kwargs,
    ):
        self.quant_method = QuantizationMethod.TEQ
        self.bits = bits
        self.tokenizer = tokenizer
        self.dataset = dataset
        self.compute_dtype = compute_dtype
        self.weight_dtype = "int4" if self.bits == 4 else "int8"
        self.scale_dtype = scale_dtype
        self.group_size = group_size
        self.sym = sym
        self.scheme = "sym" if self.sym else "asym"
        self.use_layer_wise = use_layer_wise
        self.n_samples = n_samples
        self.seq_len = seq_len
        self.folding = folding
        self.absorb_layer_dict = absorb_layer_dict
        self.quant_lm_head = quant_lm_head
        self.modules_to_not_convert = kwargs.get(
            "modules_to_not_convert", ["lm_head", "transformer.output_layer", "embed_out"]
        )
        if self.quant_lm_head:
            self.modules_to_not_convert = []
        self.device = kwargs.get("device", "auto")
        self.batch_size = kwargs.pop("batch_size", 8)

    def to_diff_dict(self) -> Dict[str, Any]:
        """Removes all attributes from config which correspond to the default config attributes
        for better readability and serializes to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        config_dict = self.to_dict()

        # get the default config dict
        default_config_dict = TeqConfig().to_dict()

        serializable_config_dict = {}

        # only serialize values that differ from the default config
        for key, value in config_dict.items():
            if value != default_config_dict[key]:
                serializable_config_dict[key] = value

        return serializable_config_dict
